import asyncio
from unittest.mock import AsyncMock, MagicMock

from httpx import Response

from bb_integrations_lib.gravitate.sd_api import GravitateSDAPI
from bb_integrations_lib.pipelines.shared.allocation_matcher.matcher import AllocationMatcher
from egg import Egg, hatch_eggs

mock_bols_and_drops_response = [
    {
        "order_number": "1103",
        "order_id": "6953f4eda8bb5f01056eaca6",
        "po": None,
        "carrier": "",
        "carrier_id": None,
        "manager": None,
        "manager_id": None,
        "last_movement_update": "2025-12-30T18:32:27.749000",
        "order_date": "2025-12-30T15:51:09Z",
        "status": "complete",
        "type": "regular",
        "drops": [
            {
                "compartment_indexes": [0],
                "before_stick_inches": 22,
                "before_stick_time": "2025-12-30T18:28:44.281000Z",
                "after_stick_inches": 47,
                "after_stick_time": "2025-12-30T18:28:52.122000Z",
                "drop_id": "6953f4eda8bb5f01056eac9e",
                "location": "1082",
                "location_id": "69090f2e5a7ba1bea579a965",
                "location_source_id": None,
                "location_source_system": "",
                "product": "87OCT 10%ETH - SUNOCO",
                "product_id": "6908f4469257414108faf071",
                "product_source_id": "FO-87-E10 | SUN",
                "product_source_system": "PDI",
                "tank_id": 1,
                "tank_source_id": None,
                "tank_source_system": None,
                "volume": 3998,
                "driver_id": "690912c7cca24d51e38ca2ea",
                "driver_username": "rstewart7468",
                "driver_extra_data": {"source_id": "1182", "source_system_id": "PDI"},
                "driver_name": "Joel Stewart",
                "driver_date": "2025-12-30T00:00:00",
                "driver_shift": "am",
                "driver_shift_id": "6952ab2e05e52270b51f3d0c",
                "tractor_number": "202",
                "trailer_number": "1609",
                "blend_codes": [],
                "site_counterparty_id": "6908f1eab4fcf6a36e6660e1",
                "site_counterparty_name": "Bazco",
                "freight_customer_id": None,
                "freight_customer_name": None,
            },
            {
                "compartment_indexes": [1],
                "before_stick_inches": 31,
                "before_stick_time": "2025-12-30T18:30:30.745000Z",
                "after_stick_inches": 78,
                "after_stick_time": "2025-12-30T18:30:34.427000Z",
                "drop_id": "6953f4eda8bb5f01056eac9f",
                "location": "1083",
                "location_id": "69090f2e5a7ba1bea579a969",
                "location_source_id": None,
                "location_source_system": "",
                "product": "87OCT 10%ETH - SUNOCO",
                "product_id": "6908f4469257414108faf071",
                "product_source_id": "FO-87-E10 | SUN",
                "product_source_system": "PDI",
                "tank_id": 3,
                "tank_source_id": None,
                "tank_source_system": None,
                "volume": 6001,
                "driver_id": "690912c7cca24d51e38ca2ea",
                "driver_username": "rstewart7468",
                "driver_extra_data": {"source_id": "1182", "source_system_id": "PDI"},
                "driver_name": "Joel Stewart",
                "driver_date": "2025-12-30T00:00:00",
                "driver_shift": "am",
                "driver_shift_id": "6952ab2e05e52270b51f3d0c",
                "tractor_number": "202",
                "trailer_number": "1609",
                "blend_codes": [],
                "site_counterparty_id": "6908f1eab4fcf6a36e6660e1",
                "site_counterparty_name": "Bazco",
                "freight_customer_id": None,
                "freight_customer_name": None,
            },
            {
                "compartment_indexes": [2],
                "before_stick_inches": 15,
                "before_stick_time": "2025-12-30T18:30:53.323000Z",
                "after_stick_inches": 26,
                "after_stick_time": "2025-12-30T18:30:57.401000Z",
                "drop_id": "6953f4eda8bb5f01056eaca0",
                "location": "1083",
                "location_id": "69090f2e5a7ba1bea579a969",
                "location_source_id": None,
                "location_source_system": "",
                "product": "93OCT 10%ETH - SUNOCO",
                "product_id": "6908f409b4fcf6a36e6660f3",
                "product_source_id": "FO-93-E10 | SUN",
                "product_source_system": "PDI",
                "tank_id": 1,
                "tank_source_id": None,
                "tank_source_system": None,
                "volume": 1003,
                "driver_id": "690912c7cca24d51e38ca2ea",
                "driver_username": "rstewart7468",
                "driver_extra_data": {"source_id": "1182", "source_system_id": "PDI"},
                "driver_name": "Joel Stewart",
                "driver_date": "2025-12-30T00:00:00",
                "driver_shift": "am",
                "driver_shift_id": "6952ab2e05e52270b51f3d0c",
                "tractor_number": "202",
                "trailer_number": "1609",
                "blend_codes": [],
                "site_counterparty_id": "6908f1eab4fcf6a36e6660e1",
                "site_counterparty_name": "Bazco",
                "freight_customer_id": None,
                "freight_customer_name": None,
            },
        ],
        "bols": [
            {
                "bol_number": "334618136",
                "bol_id": "6954191e8c10fc0f85cbda35",
                "date": "2025-12-30T18:23:06.337000Z",
                "details": [
                    {
                        "compartment_index": None,
                        "load_number": None,
                        "product": "87OCT 10%ETH - CITGO",
                        "product_id": "6908f409b4fcf6a36e6660f1",
                        "product_source_id": "FO-87-E10 | CIT",
                        "product_source_system": "PDI",
                        "net_volume": 10097,
                        "gross_volume": 9999,
                        "supplier": "Citgo Branded",
                        "supplier_id": "6908f1eab4fcf6a36e6660c8",
                        "supplier_source_id": "5002",
                        "supplier_source_system": "",
                        "contract": "",
                        "price_details": {
                            "price": None,
                            "price_id": None,
                            "curve_id": None,
                            "date": "2025-12-30T18:23:06.337000",
                            "counterparty_id": "6",
                        },
                    },
                    {
                        "compartment_index": None,
                        "load_number": None,
                        "product": "93OCT 10%ETH - CITGO",
                        "product_id": "6908f409b4fcf6a36e6660f2",
                        "product_source_id": "FO-93-E10 | CIT",
                        "product_source_system": "PDI",
                        "net_volume": 1026,
                        "gross_volume": 1003,
                        "supplier": "Citgo Branded",
                        "supplier_id": "6908f1eab4fcf6a36e6660c8",
                        "supplier_source_id": "5002",
                        "supplier_source_system": "",
                        "contract": "",
                        "price_details": {
                            "price": None,
                            "price_id": None,
                            "curve_id": None,
                            "date": "2025-12-30T18:23:06.337000",
                            "counterparty_id": "6",
                        },
                    },
                ],
                "load_or_drop": "Load",
                "location": "MARATHON -Detroit",
                "location_id": "69090ecad6ff2d863efaf07b",
                "location_source_id": "T-38-MI-3015",
                "location_source_system": "PDI",
                "driver_id": "690912c7cca24d51e38ca2ea",
                "driver_username": "rstewart7468",
                "driver_extra_data": {"source_id": "1182", "source_system_id": "PDI"},
                "driver_name": "Joel Stewart",
                "driver_date": "2025-12-30T00:00:00",
                "driver_shift": "am",
                "driver_shift_id": "6952ab2e05e52270b51f3d0c",
                "tractor_number": "202",
                "trailer_number": "1609",
            }
        ],
        "costs": [
            {
                "cost_type": "freight",
                "carrier": "missing",
                "carrier_source_id": "",
                "carrier_source_system": "",
                "per_unit_price": 0.034100000000000005,
                "cost_amount": 375.1,
            }
        ],
        "validation_bypass_on": None,
        "has_additives": False,
        "estimated_freight": {
            "id": "6953f4eda8bb5f01056eaca4",
            "parent": None,
            "created_at": "2025-12-30T15:51:09.909000",
            "type": "freight",
            "status": "open",
            "reversed": False,
            "transactions": [
                {
                    "model_type": "base",
                    "type": "Base Freight",
                    "subtype": "Band 31-35",
                    "rate": 0.0341,
                    "amount": 4000,
                    "total": 136.4,
                    "uom": "gallons",
                    "errors": [],
                    "id": "6953f4eda8bb5f01056eaca1",
                    "product_group": "Gasoline",
                    "product_id": "6908f4469257414108faf071",
                    "product_name": "87OCT 10%ETH - SUNOCO",
                    "origin": "MARATHON -Detroit",
                    "origin_id": "69090ecad6ff2d863efaf07b",
                    "origin_override": None,
                    "origin_id_override": None,
                    "destination": "1082",
                    "destination_id": "69090f2e5a7ba1bea579a967",
                    "destination_override": None,
                    "destination_id_override": None,
                    "legs": [
                        {
                            "origin": "1082",
                            "origin_id": "69090f2e5a7ba1bea579a965",
                            "destination": "1083",
                            "destination_id": "69090f2e5a7ba1bea579a969",
                            "distance": 16.05,
                        },
                        {
                            "origin": "MARATHON -Detroit",
                            "origin_id": "69090ecad6ff2d863efaf07b",
                            "destination": "1082",
                            "destination_id": "69090f2e5a7ba1bea579a965",
                            "distance": 14.37,
                        },
                    ],
                    "use_surcharge": True,
                    "manual": False,
                    "has_dependency": False,
                    "created_date": None,
                    "requires_approval": False,
                    "requires_approval_reason_code": False,
                    "is_approved": False,
                    "approved_by": None,
                    "approved_date": None,
                    "approved_reason_code": None,
                    "amount_override": None,
                    "rate_override": None,
                    "product_id_override": None,
                    "product_name_override": None,
                    "product_group_override": None,
                    "subtype_override": None,
                    "uom_override": None,
                    "total_override": None,
                    "override_by": None,
                    "override_date": None,
                    "exclude_from_invoice": False,
                    "credit_rebill_metadata": None,
                    "extra_data": {"mileage": 30},
                    "gross_volume": None,
                    "net_volume": None,
                    "bol_date": None,
                    "bol_number": None,
                    "bol_number_override": None,
                    "delivery_date": None,
                },
                {
                    "model_type": "base",
                    "type": "Base Freight",
                    "subtype": "Band 31-35",
                    "rate": 0.0341,
                    "amount": 6000,
                    "total": 204.6,
                    "uom": "gallons",
                    "errors": [],
                    "id": "6953f4eda8bb5f01056eaca2",
                    "product_group": "Gasoline",
                    "product_id": "6908f4469257414108faf071",
                    "product_name": "87OCT 10%ETH - SUNOCO",
                    "origin": "MARATHON -Detroit",
                    "origin_id": "69090ecad6ff2d863efaf07b",
                    "origin_override": None,
                    "origin_id_override": None,
                    "destination": "1083",
                    "destination_id": "69090f2e5a7ba1bea579a96b",
                    "destination_override": None,
                    "destination_id_override": None,
                    "legs": [
                        {
                            "origin": "1082",
                            "origin_id": "69090f2e5a7ba1bea579a965",
                            "destination": "1083",
                            "destination_id": "69090f2e5a7ba1bea579a969",
                            "distance": 16.05,
                        },
                        {
                            "origin": "MARATHON -Detroit",
                            "origin_id": "69090ecad6ff2d863efaf07b",
                            "destination": "1082",
                            "destination_id": "69090f2e5a7ba1bea579a965",
                            "distance": 14.37,
                        },
                    ],
                    "use_surcharge": True,
                    "manual": False,
                    "has_dependency": False,
                    "created_date": None,
                    "requires_approval": False,
                    "requires_approval_reason_code": False,
                    "is_approved": False,
                    "approved_by": None,
                    "approved_date": None,
                    "approved_reason_code": None,
                    "amount_override": None,
                    "rate_override": None,
                    "product_id_override": None,
                    "product_name_override": None,
                    "product_group_override": None,
                    "subtype_override": None,
                    "uom_override": None,
                    "total_override": None,
                    "override_by": None,
                    "override_date": None,
                    "exclude_from_invoice": False,
                    "credit_rebill_metadata": None,
                    "extra_data": {"mileage": 30},
                    "gross_volume": None,
                    "net_volume": None,
                    "bol_date": None,
                    "bol_number": None,
                    "bol_number_override": None,
                    "delivery_date": None,
                },
                {
                    "model_type": "base",
                    "type": "Base Freight",
                    "subtype": "Band 31-35",
                    "rate": 0.0341,
                    "amount": 1000,
                    "total": 34.1,
                    "uom": "gallons",
                    "errors": [],
                    "id": "6953f4eda8bb5f01056eaca3",
                    "product_group": "Gasoline",
                    "product_id": "6908f409b4fcf6a36e6660f3",
                    "product_name": "93OCT 10%ETH - SUNOCO",
                    "origin": "MARATHON -Detroit",
                    "origin_id": "69090ecad6ff2d863efaf07b",
                    "origin_override": None,
                    "origin_id_override": None,
                    "destination": "1083",
                    "destination_id": "69090f2e5a7ba1bea579a96b",
                    "destination_override": None,
                    "destination_id_override": None,
                    "legs": [
                        {
                            "origin": "1082",
                            "origin_id": "69090f2e5a7ba1bea579a965",
                            "destination": "1083",
                            "destination_id": "69090f2e5a7ba1bea579a969",
                            "distance": 16.05,
                        },
                        {
                            "origin": "MARATHON -Detroit",
                            "origin_id": "69090ecad6ff2d863efaf07b",
                            "destination": "1082",
                            "destination_id": "69090f2e5a7ba1bea579a965",
                            "distance": 14.37,
                        },
                    ],
                    "use_surcharge": True,
                    "manual": False,
                    "has_dependency": False,
                    "created_date": None,
                    "requires_approval": False,
                    "requires_approval_reason_code": False,
                    "is_approved": False,
                    "approved_by": None,
                    "approved_date": None,
                    "approved_reason_code": None,
                    "amount_override": None,
                    "rate_override": None,
                    "product_id_override": None,
                    "product_name_override": None,
                    "product_group_override": None,
                    "subtype_override": None,
                    "uom_override": None,
                    "total_override": None,
                    "override_by": None,
                    "override_date": None,
                    "exclude_from_invoice": False,
                    "credit_rebill_metadata": None,
                    "extra_data": {"mileage": 30},
                    "gross_volume": None,
                    "net_volume": None,
                    "bol_date": None,
                    "bol_number": None,
                    "bol_number_override": None,
                    "delivery_date": None,
                },
            ],
            "counterparty_id": "6908f1eab4fcf6a36e6660e1",
            "book_type": "Cost",
            "effective_date_used": "2025-12-30T00:00:00",
            "invoice_group_id": None,
            "invoice_number": None,
            "accessorial_invoice_number": None,
            "new_version_available": False,
            "new_version_total": None,
            "error": None,
        },
        "actual_freight": {
            "id": "69541abb8c10fc0f85cbda90",
            "parent": None,
            "created_at": "2025-12-30T18:32:27.707000",
            "type": "freight",
            "status": "open",
            "reversed": False,
            "transactions": [
                {
                    "model_type": "base",
                    "type": "Base Freight",
                    "subtype": "Band 31-35",
                    "rate": 0.0341,
                    "amount": 4037,
                    "total": 137.66,
                    "uom": "gallons",
                    "errors": [],
                    "id": "69541abb8c10fc0f85cbda8d",
                    "product_group": "Gasoline",
                    "product_id": "6908f4469257414108faf071",
                    "product_name": "87OCT 10%ETH - SUNOCO",
                    "origin": "MARATHON -Detroit",
                    "origin_id": "69090ecad6ff2d863efaf07b",
                    "origin_override": None,
                    "origin_id_override": None,
                    "destination": "1082",
                    "destination_id": "69090f2e5a7ba1bea579a965",
                    "destination_override": None,
                    "destination_id_override": None,
                    "legs": [
                        {
                            "origin": "1082",
                            "origin_id": "69090f2e5a7ba1bea579a965",
                            "destination": "1083",
                            "destination_id": "69090f2e5a7ba1bea579a969",
                            "distance": 16.05,
                        },
                        {
                            "origin": "MARATHON -Detroit",
                            "origin_id": "69090ecad6ff2d863efaf07b",
                            "destination": "1082",
                            "destination_id": "69090f2e5a7ba1bea579a965",
                            "distance": 14.37,
                        },
                    ],
                    "use_surcharge": True,
                    "manual": False,
                    "has_dependency": False,
                    "created_date": None,
                    "requires_approval": False,
                    "requires_approval_reason_code": False,
                    "is_approved": False,
                    "approved_by": None,
                    "approved_date": None,
                    "approved_reason_code": None,
                    "amount_override": None,
                    "rate_override": None,
                    "product_id_override": None,
                    "product_name_override": None,
                    "product_group_override": None,
                    "subtype_override": None,
                    "uom_override": None,
                    "total_override": None,
                    "override_by": None,
                    "override_date": None,
                    "exclude_from_invoice": False,
                    "credit_rebill_metadata": None,
                    "extra_data": {"mileage": 30},
                    "gross_volume": 3998,
                    "net_volume": 4037,
                    "bol_date": "2025-12-30T18:23:06.337000",
                    "bol_number": "334618136",
                    "bol_number_override": None,
                    "delivery_date": "2025-12-30T18:28:52.122000",
                },
                {
                    "model_type": "base",
                    "type": "Base Freight",
                    "subtype": "Band 31-35",
                    "rate": 0.0341,
                    "amount": 1026,
                    "total": 34.99,
                    "uom": "gallons",
                    "errors": [],
                    "id": "69541abb8c10fc0f85cbda8e",
                    "product_group": "Gasoline",
                    "product_id": "6908f409b4fcf6a36e6660f3",
                    "product_name": "93OCT 10%ETH - SUNOCO",
                    "origin": "MARATHON -Detroit",
                    "origin_id": "69090ecad6ff2d863efaf07b",
                    "origin_override": None,
                    "origin_id_override": None,
                    "destination": "1083",
                    "destination_id": "69090f2e5a7ba1bea579a969",
                    "destination_override": None,
                    "destination_id_override": None,
                    "legs": [
                        {
                            "origin": "1082",
                            "origin_id": "69090f2e5a7ba1bea579a965",
                            "destination": "1083",
                            "destination_id": "69090f2e5a7ba1bea579a969",
                            "distance": 16.05,
                        },
                        {
                            "origin": "MARATHON -Detroit",
                            "origin_id": "69090ecad6ff2d863efaf07b",
                            "destination": "1082",
                            "destination_id": "69090f2e5a7ba1bea579a965",
                            "distance": 14.37,
                        },
                    ],
                    "use_surcharge": True,
                    "manual": False,
                    "has_dependency": False,
                    "created_date": None,
                    "requires_approval": False,
                    "requires_approval_reason_code": False,
                    "is_approved": False,
                    "approved_by": None,
                    "approved_date": None,
                    "approved_reason_code": None,
                    "amount_override": None,
                    "rate_override": None,
                    "product_id_override": None,
                    "product_name_override": None,
                    "product_group_override": None,
                    "subtype_override": None,
                    "uom_override": None,
                    "total_override": None,
                    "override_by": None,
                    "override_date": None,
                    "exclude_from_invoice": False,
                    "credit_rebill_metadata": None,
                    "extra_data": {"mileage": 30},
                    "gross_volume": 1003,
                    "net_volume": 1026,
                    "bol_date": "2025-12-30T18:23:06.337000",
                    "bol_number": "334618136",
                    "bol_number_override": None,
                    "delivery_date": "2025-12-30T18:30:57.401000",
                },
                {
                    "model_type": "base",
                    "type": "Base Freight",
                    "subtype": "Band 31-35",
                    "rate": 0.0341,
                    "amount": 6060,
                    "total": 206.65,
                    "uom": "gallons",
                    "errors": [],
                    "id": "69541abb8c10fc0f85cbda8f",
                    "product_group": "Gasoline",
                    "product_id": "6908f4469257414108faf071",
                    "product_name": "87OCT 10%ETH - SUNOCO",
                    "origin": "MARATHON -Detroit",
                    "origin_id": "69090ecad6ff2d863efaf07b",
                    "origin_override": None,
                    "origin_id_override": None,
                    "destination": "1083",
                    "destination_id": "69090f2e5a7ba1bea579a969",
                    "destination_override": None,
                    "destination_id_override": None,
                    "legs": [
                        {
                            "origin": "1082",
                            "origin_id": "69090f2e5a7ba1bea579a965",
                            "destination": "1083",
                            "destination_id": "69090f2e5a7ba1bea579a969",
                            "distance": 16.05,
                        },
                        {
                            "origin": "MARATHON -Detroit",
                            "origin_id": "69090ecad6ff2d863efaf07b",
                            "destination": "1082",
                            "destination_id": "69090f2e5a7ba1bea579a965",
                            "distance": 14.37,
                        },
                    ],
                    "use_surcharge": True,
                    "manual": False,
                    "has_dependency": False,
                    "created_date": None,
                    "requires_approval": False,
                    "requires_approval_reason_code": False,
                    "is_approved": False,
                    "approved_by": None,
                    "approved_date": None,
                    "approved_reason_code": None,
                    "amount_override": None,
                    "rate_override": None,
                    "product_id_override": None,
                    "product_name_override": None,
                    "product_group_override": None,
                    "subtype_override": None,
                    "uom_override": None,
                    "total_override": None,
                    "override_by": None,
                    "override_date": None,
                    "exclude_from_invoice": False,
                    "credit_rebill_metadata": None,
                    "extra_data": {"mileage": 30},
                    "gross_volume": 6001,
                    "net_volume": 6060,
                    "bol_date": "2025-12-30T18:23:06.337000",
                    "bol_number": "334618136",
                    "bol_number_override": None,
                    "delivery_date": "2025-12-30T18:30:34.427000",
                },
            ],
            "counterparty_id": "6908f1eab4fcf6a36e6660e1",
            "book_type": "Cost",
            "effective_date_used": "2025-12-30T00:00:00",
            "invoice_group_id": None,
            "invoice_number": None,
            "accessorial_invoice_number": None,
            "new_version_available": False,
            "new_version_total": None,
            "error": None,
        },
        "allocated_bol_error": None,
        "allocated_bol_issue": None,
        "allocated_bols": [
            {
                "bol_number": "334618136",
                "bol_id": "6954191e8c10fc0f85cbda35",
                "bol_terminal_id": "69090ecad6ff2d863efaf07b",
                "bol_terminal": "MARATHON -Detroit",
                "bol_product": "87OCT 10%ETH - CITGO",
                "bol_product_id": "6908f409b4fcf6a36e6660f1",
                "bol_supplier": "Citgo Branded",
                "bol_supplier_id": "6908f1eab4fcf6a36e6660c8",
                "bol_net_volume_original": 10097,
                "bol_gross_volume_original": 9999,
                "store_number": "1082",
                "store_id": "69090f2e5a7ba1bea579a967",
                "location_id": "69090f2e5a7ba1bea579a965",
                "store_tank": 1,
                "store_product": "87OCT 10%ETH - SUNOCO",
                "store_product_id": "6908f4469257414108faf071",
                "store_timezone": "America/New_York",
                "bol_net_volume_allocated": 4037,
                "bol_gross_volume_allocated": 3998,
                "driver_schedule": "6952ab2e05e52270b51f3d0c",
                "bol_date": "2025-12-30T18:23:06.337000",
                "delivered_date": "2025-12-30T18:28:52.122000",
                "price": {
                    "contract": "",
                    "price_type": "rack",
                    "price_id": "6953080ab9ee7ece772690e7",
                    "curve_id": "6945cec2f88ce49ef5fc504a",
                    "price": 1.6712,
                    "date": "2025-12-30T18:23:06.337000",
                    "expired": False,
                },
            },
            {
                "bol_number": "334618136",
                "bol_id": "6954191e8c10fc0f85cbda35",
                "bol_terminal_id": "69090ecad6ff2d863efaf07b",
                "bol_terminal": "MARATHON -Detroit",
                "bol_product": "87OCT 10%ETH - CITGO",
                "bol_product_id": "6908f409b4fcf6a36e6660f1",
                "bol_supplier": "Citgo Branded",
                "bol_supplier_id": "6908f1eab4fcf6a36e6660c8",
                "bol_net_volume_original": 10097,
                "bol_gross_volume_original": 9999,
                "store_number": "1083",
                "store_id": "69090f2e5a7ba1bea579a96b",
                "location_id": "69090f2e5a7ba1bea579a969",
                "store_tank": 3,
                "store_product": "87OCT 10%ETH - SUNOCO",
                "store_product_id": "6908f4469257414108faf071",
                "store_timezone": "America/New_York",
                "bol_net_volume_allocated": 6060,
                "bol_gross_volume_allocated": 6001,
                "driver_schedule": "6952ab2e05e52270b51f3d0c",
                "bol_date": "2025-12-30T18:23:06.337000",
                "delivered_date": "2025-12-30T18:30:34.427000",
                "price": {
                    "contract": "",
                    "price_type": "rack",
                    "price_id": "6953080ab9ee7ece772690e7",
                    "curve_id": "6945cec2f88ce49ef5fc504a",
                    "price": 1.6712,
                    "date": "2025-12-30T18:23:06.337000",
                    "expired": False,
                },
            },
            {
                "bol_number": "334618136",
                "bol_id": "6954191e8c10fc0f85cbda35",
                "bol_terminal_id": "69090ecad6ff2d863efaf07b",
                "bol_terminal": "MARATHON -Detroit",
                "bol_product": "93OCT 10%ETH - CITGO",
                "bol_product_id": "6908f409b4fcf6a36e6660f2",
                "bol_supplier": "Citgo Branded",
                "bol_supplier_id": "6908f1eab4fcf6a36e6660c8",
                "bol_net_volume_original": 1026,
                "bol_gross_volume_original": 1003,
                "store_number": "1083",
                "store_id": "69090f2e5a7ba1bea579a96b",
                "location_id": "69090f2e5a7ba1bea579a969",
                "store_tank": 1,
                "store_product": "93OCT 10%ETH - SUNOCO",
                "store_product_id": "6908f409b4fcf6a36e6660f3",
                "store_timezone": "America/New_York",
                "bol_net_volume_allocated": 1026,
                "bol_gross_volume_allocated": 1003,
                "driver_schedule": "6952ab2e05e52270b51f3d0c",
                "bol_date": "2025-12-30T18:23:06.337000",
                "delivered_date": "2025-12-30T18:30:57.401000",
                "price": {
                    "contract": "",
                    "price_type": "rack",
                    "price_id": "6953080ab9ee7ece772690ea",
                    "curve_id": "6945cec2f88ce49ef5fc5050",
                    "price": 2.6813,
                    "date": "2025-12-30T18:23:06.337000",
                    "expired": False,
                },
            },
        ],
        "last_change_date": "2025-12-30T18:32:27.749000",
        "reference_order_number": None,
    },
    {
        "order_number": "1196",
        "order_id": "69581c65ee5aaa626ed7ddee",
        "po": None,
        "carrier": "",
        "carrier_id": None,
        "manager": None,
        "manager_id": None,
        "last_movement_update": "2026-01-02T19:29:59.497000",
        "order_date": "2026-01-02T19:28:37Z",
        "status": "complete",
        "type": "regular",
        "drops": [
            {
                "compartment_indexes": [
                    0
                ],
                "before_stick_inches": None,
                "before_stick_time": "2026-01-02T18:29:00Z",
                "after_stick_inches": None,
                "after_stick_time": "2026-01-02T18:29:00Z",
                "drop_id": "69581c65ee5aaa626ed7dde6",
                "location": "1082",
                "location_id": "69090f2e5a7ba1bea579a965",
                "location_source_id": None,
                "location_source_system": "",
                "product": "87OCT 10%ETH - SUNOCO",
                "product_id": "6908f4469257414108faf071",
                "product_source_id": "FO-87-E10 | SUN",
                "product_source_system": "PDI",
                "tank_id": 1,
                "tank_source_id": None,
                "tank_source_system": None,
                "volume": 4000,
                "driver_id": None,
                "driver_username": None,
                "driver_extra_data": None,
                "driver_name": None,
                "driver_date": None,
                "driver_shift": None,
                "driver_shift_id": None,
                "tractor_number": None,
                "trailer_number": None,
                "blend_codes": [],
                "site_counterparty_id": "6908f1eab4fcf6a36e6660e1",
                "site_counterparty_name": "Bazco",
                "freight_customer_id": None,
                "freight_customer_name": None
            },
            {
                "compartment_indexes": [
                    1
                ],
                "before_stick_inches": None,
                "before_stick_time": "2026-01-02T18:29:00Z",
                "after_stick_inches": None,
                "after_stick_time": "2026-01-02T18:29:00Z",
                "drop_id": "69581c65ee5aaa626ed7dde7",
                "location": "1083",
                "location_id": "69090f2e5a7ba1bea579a969",
                "location_source_id": None,
                "location_source_system": "",
                "product": "87OCT 10%ETH - SUNOCO",
                "product_id": "6908f4469257414108faf071",
                "product_source_id": "FO-87-E10 | SUN",
                "product_source_system": "PDI",
                "tank_id": 3,
                "tank_source_id": None,
                "tank_source_system": None,
                "volume": 6000,
                "driver_id": None,
                "driver_username": None,
                "driver_extra_data": None,
                "driver_name": None,
                "driver_date": None,
                "driver_shift": None,
                "driver_shift_id": None,
                "tractor_number": None,
                "trailer_number": None,
                "blend_codes": [],
                "site_counterparty_id": "6908f1eab4fcf6a36e6660e1",
                "site_counterparty_name": "Bazco",
                "freight_customer_id": None,
                "freight_customer_name": None
            },
            {
                "compartment_indexes": [
                    2
                ],
                "before_stick_inches": None,
                "before_stick_time": "2026-01-02T18:29:00Z",
                "after_stick_inches": None,
                "after_stick_time": "2026-01-02T18:29:00Z",
                "drop_id": "69581c65ee5aaa626ed7dde8",
                "location": "1083",
                "location_id": "69090f2e5a7ba1bea579a969",
                "location_source_id": None,
                "location_source_system": "",
                "product": "93OCT 10%ETH - SUNOCO",
                "product_id": "6908f409b4fcf6a36e6660f3",
                "product_source_id": "FO-93-E10 | SUN",
                "product_source_system": "PDI",
                "tank_id": 1,
                "tank_source_id": None,
                "tank_source_system": None,
                "volume": 1000,
                "driver_id": None,
                "driver_username": None,
                "driver_extra_data": None,
                "driver_name": None,
                "driver_date": None,
                "driver_shift": None,
                "driver_shift_id": None,
                "tractor_number": None,
                "trailer_number": None,
                "blend_codes": [],
                "site_counterparty_id": "6908f1eab4fcf6a36e6660e1",
                "site_counterparty_name": "Bazco",
                "freight_customer_id": None,
                "freight_customer_name": None
            }
        ],
        "bols": [
            {
                "bol_number": "243242",
                "bol_id": "69581c9d37fc1998fed7e946",
                "date": "2026-01-02T18:28:00Z",
                "details": [
                    {
                        "compartment_index": None,
                        "load_number": None,
                        "product": "87OCT 10%ETH - SUNOCO",
                        "product_id": "6908f4469257414108faf071",
                        "product_source_id": "FO-87-E10 | SUN",
                        "product_source_system": "PDI",
                        "net_volume": 10000,
                        "gross_volume": 9999,
                        "supplier": "Sunoco",
                        "supplier_id": "6908f1eab4fcf6a36e6660d9",
                        "supplier_source_id": "5008",
                        "supplier_source_system": "",
                        "contract": "",
                        "price_details": {
                            "price": None,
                            "price_id": None,
                            "curve_id": None,
                            "date": "2026-01-02T18:28:00",
                            "counterparty_id": "6"
                        }
                    },
                    {
                        "compartment_index": None,
                        "load_number": None,
                        "product": "93OCT 10%ETH - SUNOCO",
                        "product_id": "6908f409b4fcf6a36e6660f3",
                        "product_source_id": "FO-93-E10 | SUN",
                        "product_source_system": "PDI",
                        "net_volume": 1026,
                        "gross_volume": 1003,
                        "supplier": "Sunoco",
                        "supplier_id": "6908f1eab4fcf6a36e6660d9",
                        "supplier_source_id": "5008",
                        "supplier_source_system": "",
                        "contract": "",
                        "price_details": {
                            "price": None,
                            "price_id": None,
                            "curve_id": None,
                            "date": "2026-01-02T18:28:00",
                            "counterparty_id": "6"
                        }
                    }
                ],
                "load_or_drop": "Load",
                "location": "SUNOCO -RIVER ROUGE",
                "location_id": "69090ecad6ff2d863efaf084",
                "location_source_id": "T-38-MI-3030",
                "location_source_system": "PDI",
                "driver_id": None,
                "driver_username": None,
                "driver_extra_data": None,
                "driver_name": None,
                "driver_date": None,
                "driver_shift": None,
                "driver_shift_id": None,
                "tractor_number": None,
                "trailer_number": None
            }
        ],
        "costs": [
            {
                "cost_type": "freight",
                "carrier": "missing",
                "carrier_source_id": "",
                "carrier_source_system": "",
                "per_unit_price": 0.0324,
                "cost_amount": 356.4
            }
        ],
        "validation_bypass_on": "2026-01-02T19:29:59.303000",
        "has_additives": False,
        "estimated_freight": {
            "id": "69581c65ee5aaa626ed7ddec",
            "parent": None,
            "created_at": "2026-01-02T19:28:37.447000",
            "type": "freight",
            "status": "open",
            "reversed": False,
            "transactions": [
                {
                    "model_type": "base",
                    "type": "Base Freight",
                    "subtype": "Band 26-30",
                    "rate": 0.0324,
                    "amount": 4000,
                    "total": 129.6,
                    "uom": "gallons",
                    "errors": [],
                    "id": "69581c65ee5aaa626ed7dde9",
                    "product_group": "Gasoline",
                    "product_id": "6908f4469257414108faf071",
                    "product_name": "87OCT 10%ETH - SUNOCO",
                    "origin": "SUNOCO -RIVER ROUGE",
                    "origin_id": "69090ecad6ff2d863efaf084",
                    "origin_override": None,
                    "origin_id_override": None,
                    "destination": "1082",
                    "destination_id": "69090f2e5a7ba1bea579a967",
                    "destination_override": None,
                    "destination_id_override": None,
                    "legs": [
                        {
                            "origin": "SUNOCO -RIVER ROUGE",
                            "origin_id": "69090ecad6ff2d863efaf084",
                            "destination": "1082",
                            "destination_id": "69090f2e5a7ba1bea579a965",
                            "distance": 13.6
                        },
                        {
                            "origin": "1082",
                            "origin_id": "69090f2e5a7ba1bea579a965",
                            "destination": "1083",
                            "destination_id": "69090f2e5a7ba1bea579a969",
                            "distance": 16.05
                        }
                    ],
                    "use_surcharge": True,
                    "manual": False,
                    "has_dependency": False,
                    "created_date": None,
                    "requires_approval": False,
                    "requires_approval_reason_code": False,
                    "is_approved": False,
                    "approved_by": None,
                    "approved_date": None,
                    "approved_reason_code": None,
                    "amount_override": None,
                    "rate_override": None,
                    "product_id_override": None,
                    "product_name_override": None,
                    "product_group_override": None,
                    "subtype_override": None,
                    "uom_override": None,
                    "total_override": None,
                    "override_by": None,
                    "override_date": None,
                    "exclude_from_invoice": False,
                    "credit_rebill_metadata": None,
                    "extra_data": {
                        "mileage": 30
                    },
                    "gross_volume": None,
                    "net_volume": None,
                    "bol_date": None,
                    "bol_number": None,
                    "bol_number_override": None,
                    "delivery_date": None
                },
                {
                    "model_type": "base",
                    "type": "Base Freight",
                    "subtype": "Band 26-30",
                    "rate": 0.0324,
                    "amount": 6000,
                    "total": 194.4,
                    "uom": "gallons",
                    "errors": [],
                    "id": "69581c65ee5aaa626ed7ddea",
                    "product_group": "Gasoline",
                    "product_id": "6908f4469257414108faf071",
                    "product_name": "87OCT 10%ETH - SUNOCO",
                    "origin": "SUNOCO -RIVER ROUGE",
                    "origin_id": "69090ecad6ff2d863efaf084",
                    "origin_override": None,
                    "origin_id_override": None,
                    "destination": "1083",
                    "destination_id": "69090f2e5a7ba1bea579a96b",
                    "destination_override": None,
                    "destination_id_override": None,
                    "legs": [
                        {
                            "origin": "SUNOCO -RIVER ROUGE",
                            "origin_id": "69090ecad6ff2d863efaf084",
                            "destination": "1082",
                            "destination_id": "69090f2e5a7ba1bea579a965",
                            "distance": 13.6
                        },
                        {
                            "origin": "1082",
                            "origin_id": "69090f2e5a7ba1bea579a965",
                            "destination": "1083",
                            "destination_id": "69090f2e5a7ba1bea579a969",
                            "distance": 16.05
                        }
                    ],
                    "use_surcharge": True,
                    "manual": False,
                    "has_dependency": False,
                    "created_date": None,
                    "requires_approval": False,
                    "requires_approval_reason_code": False,
                    "is_approved": False,
                    "approved_by": None,
                    "approved_date": None,
                    "approved_reason_code": None,
                    "amount_override": None,
                    "rate_override": None,
                    "product_id_override": None,
                    "product_name_override": None,
                    "product_group_override": None,
                    "subtype_override": None,
                    "uom_override": None,
                    "total_override": None,
                    "override_by": None,
                    "override_date": None,
                    "exclude_from_invoice": False,
                    "credit_rebill_metadata": None,
                    "extra_data": {
                        "mileage": 30
                    },
                    "gross_volume": None,
                    "net_volume": None,
                    "bol_date": None,
                    "bol_number": None,
                    "bol_number_override": None,
                    "delivery_date": None
                },
                {
                    "model_type": "base",
                    "type": "Base Freight",
                    "subtype": "Band 26-30",
                    "rate": 0.0324,
                    "amount": 1000,
                    "total": 32.4,
                    "uom": "gallons",
                    "errors": [],
                    "id": "69581c65ee5aaa626ed7ddeb",
                    "product_group": "Gasoline",
                    "product_id": "6908f409b4fcf6a36e6660f3",
                    "product_name": "93OCT 10%ETH - SUNOCO",
                    "origin": "SUNOCO -RIVER ROUGE",
                    "origin_id": "69090ecad6ff2d863efaf084",
                    "origin_override": None,
                    "origin_id_override": None,
                    "destination": "1083",
                    "destination_id": "69090f2e5a7ba1bea579a96b",
                    "destination_override": None,
                    "destination_id_override": None,
                    "legs": [
                        {
                            "origin": "SUNOCO -RIVER ROUGE",
                            "origin_id": "69090ecad6ff2d863efaf084",
                            "destination": "1082",
                            "destination_id": "69090f2e5a7ba1bea579a965",
                            "distance": 13.6
                        },
                        {
                            "origin": "1082",
                            "origin_id": "69090f2e5a7ba1bea579a965",
                            "destination": "1083",
                            "destination_id": "69090f2e5a7ba1bea579a969",
                            "distance": 16.05
                        }
                    ],
                    "use_surcharge": True,
                    "manual": False,
                    "has_dependency": False,
                    "created_date": None,
                    "requires_approval": False,
                    "requires_approval_reason_code": False,
                    "is_approved": False,
                    "approved_by": None,
                    "approved_date": None,
                    "approved_reason_code": None,
                    "amount_override": None,
                    "rate_override": None,
                    "product_id_override": None,
                    "product_name_override": None,
                    "product_group_override": None,
                    "subtype_override": None,
                    "uom_override": None,
                    "total_override": None,
                    "override_by": None,
                    "override_date": None,
                    "exclude_from_invoice": False,
                    "credit_rebill_metadata": None,
                    "extra_data": {
                        "mileage": 30
                    },
                    "gross_volume": None,
                    "net_volume": None,
                    "bol_date": None,
                    "bol_number": None,
                    "bol_number_override": None,
                    "delivery_date": None
                }
            ],
            "counterparty_id": "6908f1eab4fcf6a36e6660e1",
            "book_type": "Cost",
            "effective_date_used": "2026-01-03T00:00:00",
            "invoice_group_id": None,
            "invoice_number": None,
            "accessorial_invoice_number": None,
            "new_version_available": False,
            "new_version_total": None,
            "error": None
        },
        "actual_freight": {
            "id": "69581cb74309aa3937d7df5b",
            "parent": None,
            "created_at": "2026-01-02T19:29:59.488000",
            "type": "freight",
            "status": "open",
            "reversed": False,
            "transactions": [
                {
                    "model_type": "base",
                    "type": "Base Freight",
                    "subtype": "Band 26-30",
                    "rate": 0.0324,
                    "amount": 4000,
                    "total": 129.6,
                    "uom": "gallons",
                    "errors": [],
                    "id": "69581cb74309aa3937d7df56",
                    "product_group": "Gasoline",
                    "product_id": "6908f4469257414108faf071",
                    "product_name": "87OCT 10%ETH - SUNOCO",
                    "origin": "SUNOCO -RIVER ROUGE",
                    "origin_id": "69090ecad6ff2d863efaf084",
                    "origin_override": None,
                    "origin_id_override": None,
                    "destination": "1082",
                    "destination_id": "69090f2e5a7ba1bea579a965",
                    "destination_override": None,
                    "destination_id_override": None,
                    "legs": [
                        {
                            "origin": "SUNOCO -RIVER ROUGE",
                            "origin_id": "69090ecad6ff2d863efaf084",
                            "destination": "1082",
                            "destination_id": "69090f2e5a7ba1bea579a965",
                            "distance": 13.6
                        },
                        {
                            "origin": "1082",
                            "origin_id": "69090f2e5a7ba1bea579a965",
                            "destination": "1083",
                            "destination_id": "69090f2e5a7ba1bea579a969",
                            "distance": 16.05
                        }
                    ],
                    "use_surcharge": True,
                    "manual": False,
                    "has_dependency": False,
                    "created_date": None,
                    "requires_approval": False,
                    "requires_approval_reason_code": False,
                    "is_approved": False,
                    "approved_by": None,
                    "approved_date": None,
                    "approved_reason_code": None,
                    "amount_override": None,
                    "rate_override": None,
                    "product_id_override": None,
                    "product_name_override": None,
                    "product_group_override": None,
                    "subtype_override": None,
                    "uom_override": None,
                    "total_override": None,
                    "override_by": None,
                    "override_date": None,
                    "exclude_from_invoice": False,
                    "credit_rebill_metadata": None,
                    "extra_data": {
                        "mileage": 30
                    },
                    "gross_volume": 4000,
                    "net_volume": 4000,
                    "bol_date": "2026-01-02T18:28:00",
                    "bol_number": "243242",
                    "bol_number_override": None,
                    "delivery_date": "2026-01-02T18:29:00"
                },
                {
                    "model_type": "base",
                    "type": "Base Freight",
                    "subtype": "Band 26-30",
                    "rate": 0.0324,
                    "amount": 1026,
                    "total": 33.24,
                    "uom": "gallons",
                    "errors": [],
                    "id": "69581cb74309aa3937d7df58",
                    "product_group": "Gasoline",
                    "product_id": "6908f409b4fcf6a36e6660f3",
                    "product_name": "93OCT 10%ETH - SUNOCO",
                    "origin": "SUNOCO -RIVER ROUGE",
                    "origin_id": "69090ecad6ff2d863efaf084",
                    "origin_override": None,
                    "origin_id_override": None,
                    "destination": "1083",
                    "destination_id": "69090f2e5a7ba1bea579a969",
                    "destination_override": None,
                    "destination_id_override": None,
                    "legs": [
                        {
                            "origin": "SUNOCO -RIVER ROUGE",
                            "origin_id": "69090ecad6ff2d863efaf084",
                            "destination": "1082",
                            "destination_id": "69090f2e5a7ba1bea579a965",
                            "distance": 13.6
                        },
                        {
                            "origin": "1082",
                            "origin_id": "69090f2e5a7ba1bea579a965",
                            "destination": "1083",
                            "destination_id": "69090f2e5a7ba1bea579a969",
                            "distance": 16.05
                        }
                    ],
                    "use_surcharge": True,
                    "manual": False,
                    "has_dependency": False,
                    "created_date": None,
                    "requires_approval": False,
                    "requires_approval_reason_code": False,
                    "is_approved": False,
                    "approved_by": None,
                    "approved_date": None,
                    "approved_reason_code": None,
                    "amount_override": None,
                    "rate_override": None,
                    "product_id_override": None,
                    "product_name_override": None,
                    "product_group_override": None,
                    "subtype_override": None,
                    "uom_override": None,
                    "total_override": None,
                    "override_by": None,
                    "override_date": None,
                    "exclude_from_invoice": False,
                    "credit_rebill_metadata": None,
                    "extra_data": {
                        "mileage": 30
                    },
                    "gross_volume": 1003,
                    "net_volume": 1026,
                    "bol_date": "2026-01-02T18:28:00",
                    "bol_number": "243242",
                    "bol_number_override": None,
                    "delivery_date": "2026-01-02T18:29:00"
                },
                {
                    "model_type": "base",
                    "type": "Base Freight",
                    "subtype": "Band 26-30",
                    "rate": 0.0324,
                    "amount": 6000,
                    "total": 194.4,
                    "uom": "gallons",
                    "errors": [],
                    "id": "69581cb74309aa3937d7df5a",
                    "product_group": "Gasoline",
                    "product_id": "6908f4469257414108faf071",
                    "product_name": "87OCT 10%ETH - SUNOCO",
                    "origin": "SUNOCO -RIVER ROUGE",
                    "origin_id": "69090ecad6ff2d863efaf084",
                    "origin_override": None,
                    "origin_id_override": None,
                    "destination": "1083",
                    "destination_id": "69090f2e5a7ba1bea579a969",
                    "destination_override": None,
                    "destination_id_override": None,
                    "legs": [
                        {
                            "origin": "SUNOCO -RIVER ROUGE",
                            "origin_id": "69090ecad6ff2d863efaf084",
                            "destination": "1082",
                            "destination_id": "69090f2e5a7ba1bea579a965",
                            "distance": 13.6
                        },
                        {
                            "origin": "1082",
                            "origin_id": "69090f2e5a7ba1bea579a965",
                            "destination": "1083",
                            "destination_id": "69090f2e5a7ba1bea579a969",
                            "distance": 16.05
                        }
                    ],
                    "use_surcharge": True,
                    "manual": False,
                    "has_dependency": False,
                    "created_date": None,
                    "requires_approval": False,
                    "requires_approval_reason_code": False,
                    "is_approved": False,
                    "approved_by": None,
                    "approved_date": None,
                    "approved_reason_code": None,
                    "amount_override": None,
                    "rate_override": None,
                    "product_id_override": None,
                    "product_name_override": None,
                    "product_group_override": None,
                    "subtype_override": None,
                    "uom_override": None,
                    "total_override": None,
                    "override_by": None,
                    "override_date": None,
                    "exclude_from_invoice": False,
                    "credit_rebill_metadata": None,
                    "extra_data": {
                        "mileage": 30
                    },
                    "gross_volume": 5999,
                    "net_volume": 6000,
                    "bol_date": "2026-01-02T18:28:00",
                    "bol_number": "243242",
                    "bol_number_override": None,
                    "delivery_date": "2026-01-02T18:29:00"
                }
            ],
            "counterparty_id": "6908f1eab4fcf6a36e6660e1",
            "book_type": "Cost",
            "effective_date_used": "2026-01-02T00:00:00",
            "invoice_group_id": None,
            "invoice_number": None,
            "accessorial_invoice_number": None,
            "new_version_available": False,
            "new_version_total": None,
            "error": None
        },
        "allocated_bol_error": None,
        "allocated_bol_issue": None,
        "allocated_bols": [
            {
                "bol_number": "243242",
                "bol_id": "69581c9d37fc1998fed7e946",
                "bol_terminal_id": "69090ecad6ff2d863efaf084",
                "bol_terminal": "SUNOCO -RIVER ROUGE",
                "bol_product": "87OCT 10%ETH - SUNOCO",
                "bol_product_id": "6908f4469257414108faf071",
                "bol_supplier": "Sunoco",
                "bol_supplier_id": "6908f1eab4fcf6a36e6660d9",
                "bol_net_volume_original": 10000,
                "bol_gross_volume_original": 9999,
                "store_number": "1082",
                "store_id": "69090f2e5a7ba1bea579a967",
                "location_id": "69090f2e5a7ba1bea579a965",
                "store_tank": 1,
                "store_product": "87OCT 10%ETH - SUNOCO",
                "store_product_id": "6908f4469257414108faf071",
                "store_timezone": "America/New_York",
                "bol_net_volume_allocated": 4000,
                "bol_gross_volume_allocated": 4000,
                "driver_schedule": None,
                "bol_date": "2026-01-02T18:28:00",
                "delivered_date": "2026-01-02T18:29:00",
                "price": {
                    "contract": "",
                    "price_type": "rack",
                    "price_id": "6955ab1fb816b69ba5c18eca",
                    "curve_id": "6945cec2f88ce49ef5fc509c",
                    "price": 1.7556,
                    "date": "2026-01-02T18:28:00",
                    "expired": True
                }
            },
            {
                "bol_number": "243242",
                "bol_id": "69581c9d37fc1998fed7e946",
                "bol_terminal_id": "69090ecad6ff2d863efaf084",
                "bol_terminal": "SUNOCO -RIVER ROUGE",
                "bol_product": "87OCT 10%ETH - SUNOCO",
                "bol_product_id": "6908f4469257414108faf071",
                "bol_supplier": "Sunoco",
                "bol_supplier_id": "6908f1eab4fcf6a36e6660d9",
                "bol_net_volume_original": 10000,
                "bol_gross_volume_original": 9999,
                "store_number": "1083",
                "store_id": "69090f2e5a7ba1bea579a96b",
                "location_id": "69090f2e5a7ba1bea579a969",
                "store_tank": 3,
                "store_product": "87OCT 10%ETH - SUNOCO",
                "store_product_id": "6908f4469257414108faf071",
                "store_timezone": "America/New_York",
                "bol_net_volume_allocated": 6000,
                "bol_gross_volume_allocated": 5999,
                "driver_schedule": None,
                "bol_date": "2026-01-02T18:28:00",
                "delivered_date": "2026-01-02T18:29:00",
                "price": {
                    "contract": "",
                    "price_type": "rack",
                    "price_id": "6955ab1fb816b69ba5c18eca",
                    "curve_id": "6945cec2f88ce49ef5fc509c",
                    "price": 1.7556,
                    "date": "2026-01-02T18:28:00",
                    "expired": True
                }
            },
            {
                "bol_number": "243242",
                "bol_id": "69581c9d37fc1998fed7e946",
                "bol_terminal_id": "69090ecad6ff2d863efaf084",
                "bol_terminal": "SUNOCO -RIVER ROUGE",
                "bol_product": "93OCT 10%ETH - SUNOCO",
                "bol_product_id": "6908f409b4fcf6a36e6660f3",
                "bol_supplier": "Sunoco",
                "bol_supplier_id": "6908f1eab4fcf6a36e6660d9",
                "bol_net_volume_original": 1026,
                "bol_gross_volume_original": 1003,
                "store_number": "1083",
                "store_id": "69090f2e5a7ba1bea579a96b",
                "location_id": "69090f2e5a7ba1bea579a969",
                "store_tank": 1,
                "store_product": "93OCT 10%ETH - SUNOCO",
                "store_product_id": "6908f409b4fcf6a36e6660f3",
                "store_timezone": "America/New_York",
                "bol_net_volume_allocated": 1026,
                "bol_gross_volume_allocated": 1003,
                "driver_schedule": None,
                "bol_date": "2026-01-02T18:28:00",
                "delivered_date": "2026-01-02T18:29:00",
                "price": {
                    "contract": "",
                    "price_type": "rack",
                    "price_id": "6955ab1fb816b69ba5c18ea9",
                    "curve_id": "6945cec2f88ce49ef5fc505a",
                    "price": 2.8061,
                    "date": "2026-01-02T18:28:00",
                    "expired": True
                }
            }
        ],
        "last_change_date": "2026-01-02T19:29:59.497000",
        "reference_order_number": None
    }
]

mock_get_orders_response = [
    {
        "type": "regular",
        "order_id": "6953f4eda8bb5f01056eaca6",
        "drivers": [
            {
                "driver_id": "690912c7cca24d51e38ca2ea",
                "driver_schedule_id": "6952ab2e05e52270b51f3d0c",
                "name": "Joel Stewart",
                "username": "rstewart7468"
            }
        ],
        "order_number": 1103,
        "order_date": "2025-12-30T15:51:09Z",
        "order_state": "complete",
        "carrier_window_start": None,
        "carrier_window_end": None,
        "carrier_window_start_local": None,
        "carrier_window_end_local": None,
        "carrier_notify_state": "none",
        "load_window_start": "2025-12-29T23:00:00Z",
        "load_window_end": "2026-01-04T23:00:00Z",
        "dispatch_window_start": "2025-12-25T23:29:00Z",
        "dispatch_window_end": "2025-12-25T23:29:00Z",
        "hauler_counterparty_id": "None",
        "hauler_counterparty_name": "",
        "hauler_counterparty_source_id": None,
        "hauler_counterparty_source_system": "",
        "hauled_by_updated_by": "JBerninger",
        "hauled_by_updated": "2025-12-30T15:51:33.218000",
        "loads": [
            {
                "location_id": "69090ecad6ff2d863efaf07b",
                "location_name": "MARATHON -Detroit",
                "location_source_id": "T-38-MI-3015",
                "location_source_system": "PDI",
                "location_type": "Terminal",
                "details": [
                    {
                        "detail_id": "6953f4eda8bb5f01056eac9c",
                        "counterparty_id": "6908f1eab4fcf6a36e6660c8",
                        "counterparty_name": "Citgo Branded",
                        "counterparty_source_id": "5002",
                        "counterparty_source_system": None,
                        "price_type": "rack",
                        "contract": "",
                        "product_id": "6908f409b4fcf6a36e6660f1",
                        "product_name": "87OCT 10%ETH - CITGO",
                        "product_source_id": "FO-87-E10 | CIT",
                        "product_source_system": "PDI",
                        "drop_product_id": "6908f4469257414108faf071",
                        "drop_product_name": "87OCT 10%ETH - SUNOCO",
                        "drop_product_source_id": "FO-87-E10 | SUN",
                        "drop_product_source_system": "PDI",
                        "quantity": 4000,
                        "compartment_index": 0,
                        "blend_code": None,
                        "product_codes": [],
                        "load_number": None
                    },
                    {
                        "detail_id": "6953f4eda8bb5f01056eac9c",
                        "counterparty_id": "6908f1eab4fcf6a36e6660c8",
                        "counterparty_name": "Citgo Branded",
                        "counterparty_source_id": "5002",
                        "counterparty_source_system": None,
                        "price_type": "rack",
                        "contract": "",
                        "product_id": "6908f409b4fcf6a36e6660f1",
                        "product_name": "87OCT 10%ETH - CITGO",
                        "product_source_id": "FO-87-E10 | CIT",
                        "product_source_system": "PDI",
                        "drop_product_id": "6908f4469257414108faf071",
                        "drop_product_name": "87OCT 10%ETH - SUNOCO",
                        "drop_product_source_id": "FO-87-E10 | SUN",
                        "drop_product_source_system": "PDI",
                        "quantity": 6000,
                        "compartment_index": 1,
                        "blend_code": None,
                        "product_codes": [],
                        "load_number": None
                    },
                    {
                        "detail_id": "6953f4eda8bb5f01056eac9d",
                        "counterparty_id": "6908f1eab4fcf6a36e6660c8",
                        "counterparty_name": "Citgo Branded",
                        "counterparty_source_id": "5002",
                        "counterparty_source_system": None,
                        "price_type": "rack",
                        "contract": "",
                        "product_id": "6908f409b4fcf6a36e6660f2",
                        "product_name": "93OCT 10%ETH - CITGO",
                        "product_source_id": "FO-93-E10 | CIT",
                        "product_source_system": "PDI",
                        "drop_product_id": "6908f409b4fcf6a36e6660f3",
                        "drop_product_name": "93OCT 10%ETH - SUNOCO",
                        "drop_product_source_id": "FO-93-E10 | SUN",
                        "drop_product_source_system": "PDI",
                        "quantity": 1000,
                        "compartment_index": 2,
                        "blend_code": None,
                        "product_codes": [],
                        "load_number": None
                    }
                ],
                "eta": "2025-12-30T18:22:55.199000",
                "status": "complete"
            }
        ],
        "drops": [
            {
                "location_id": "69090f2e5a7ba1bea579a965",
                "location_name": "1082",
                "location_source_id": None,
                "location_source_system": "",
                "location_type": "Store",
                "details": [
                    {
                        "detail_id": "6953f4eda8bb5f01056eac9e",
                        "counterparty_id": "",
                        "counterparty_name": "",
                        "counterparty_source_id": None,
                        "counterparty_source_system": None,
                        "product_id": "6908f4469257414108faf071",
                        "product_name": "87OCT 10%ETH - SUNOCO",
                        "product_source_id": "FO-87-E10 | SUN",
                        "product_source_system": "PDI",
                        "quantity": 4000,
                        "tank_id": 1,
                        "sources": [
                            {
                                "compartment_index": 0,
                                "volume": 4000
                            }
                        ],
                        "destination_codes": [],
                        "blend_codes": []
                    }
                ],
                "eta": "2025-12-30T18:26:36.339000",
                "status": "complete"
            },
            {
                "location_id": "69090f2e5a7ba1bea579a969",
                "location_name": "1083",
                "location_source_id": None,
                "location_source_system": "",
                "location_type": "Store",
                "details": [
                    {
                        "detail_id": "6953f4eda8bb5f01056eac9f",
                        "counterparty_id": "",
                        "counterparty_name": "",
                        "counterparty_source_id": None,
                        "counterparty_source_system": None,
                        "product_id": "6908f4469257414108faf071",
                        "product_name": "87OCT 10%ETH - SUNOCO",
                        "product_source_id": "FO-87-E10 | SUN",
                        "product_source_system": "PDI",
                        "quantity": 6000,
                        "tank_id": 3,
                        "sources": [
                            {
                                "compartment_index": 1,
                                "volume": 6000
                            }
                        ],
                        "destination_codes": [],
                        "blend_codes": []
                    },
                    {
                        "detail_id": "6953f4eda8bb5f01056eaca0",
                        "counterparty_id": "",
                        "counterparty_name": "",
                        "counterparty_source_id": None,
                        "counterparty_source_system": None,
                        "product_id": "6908f409b4fcf6a36e6660f3",
                        "product_name": "93OCT 10%ETH - SUNOCO",
                        "product_source_id": "FO-93-E10 | SUN",
                        "product_source_system": "PDI",
                        "quantity": 1000,
                        "tank_id": 1,
                        "sources": [
                            {
                                "compartment_index": 2,
                                "volume": 1000
                            }
                        ],
                        "destination_codes": [],
                        "blend_codes": []
                    }
                ],
                "eta": "2025-12-30T18:30:06.668000",
                "status": "complete"
            }
        ],
        "detours": [],
        "trip_status": "complete",
        "last_change_date": "2025-12-30T18:32:27.749000",
        "market": "Michigan",
        "supply_option": {
            "option": "recommended",
            "carrier_id": None,
            "carrier": None,
            "loaded_miles": 30.42233356032,
            "freight": 0.034100000000000005,
            "freight_cost": 0,
            "stops": 3,
            "prices": [
                {
                    "product_name": "87OCT 10%ETH - SUNOCO",
                    "volume": 10000,
                    "components": [
                        {
                            "timezone_data": None,
                            "curve_id": "6945cec2f88ce49ef5fc504a",
                            "price_id": "6953080ab9ee7ece772690e7",
                            "terminal_id": "69090ecad6ff2d863efaf07b",
                            "terminal": "MARATHON -Detroit",
                            "supplier_id": "6908f1eab4fcf6a36e6660c8",
                            "supplier": "Citgo Branded",
                            "price_type": "rack",
                            "product_id": "6908f409b4fcf6a36e6660f1",
                            "product": "87OCT 10%ETH - CITGO",
                            "supply_owner_id": "6908f1eab4fcf6a36e6660e1",
                            "supply_owner": "Bazco",
                            "price": 1.6712,
                            "contract": "",
                            "volume": 10000,
                            "effective_from": "2025-12-29T23:00:00Z",
                            "effective_to": "2026-01-04T23:00:00Z",
                            "group_id": None,
                            "group_effective_cutover": None,
                            "group_effective_cutover_timezone": None,
                            "group_identifier": None,
                            "group_name": None,
                            "min_constraint": None,
                            "max_constraint": None,
                            "load_number": None,
                            "load_numbers": [],
                            "directive_info": None
                        }
                    ],
                    "product_total": 16712,
                    "blend_code": None
                },
                {
                    "product_name": "93OCT 10%ETH - SUNOCO",
                    "volume": 1000,
                    "components": [
                        {
                            "timezone_data": None,
                            "curve_id": "6945cec2f88ce49ef5fc5050",
                            "price_id": "6953080ab9ee7ece772690ea",
                            "terminal_id": "69090ecad6ff2d863efaf07b",
                            "terminal": "MARATHON -Detroit",
                            "supplier_id": "6908f1eab4fcf6a36e6660c8",
                            "supplier": "Citgo Branded",
                            "price_type": "rack",
                            "product_id": "6908f409b4fcf6a36e6660f2",
                            "product": "93OCT 10%ETH - CITGO",
                            "supply_owner_id": "6908f1eab4fcf6a36e6660e1",
                            "supply_owner": "Bazco",
                            "price": 2.6813,
                            "contract": "",
                            "volume": 1000,
                            "effective_from": "2025-12-29T23:00:00Z",
                            "effective_to": "2026-01-04T23:00:00Z",
                            "group_id": None,
                            "group_effective_cutover": None,
                            "group_effective_cutover_timezone": None,
                            "group_identifier": None,
                            "group_name": None,
                            "min_constraint": None,
                            "max_constraint": None,
                            "load_number": None,
                            "load_numbers": [],
                            "directive_info": None
                        }
                    ],
                    "product_total": 2681.2999999999997,
                    "blend_code": None
                }
            ],
            "product_total": 19393.3,
            "landed_cost": 19768.399999999998,
            "supply_owner_map": {
                "definitions": [
                    {
                        "location_name": "1082",
                        "product": "87OCT 10%ETH - SUNOCO",
                        "supply_owner_id": "6908f1eab4fcf6a36e6660e1"
                    },
                    {
                        "location_name": "1083",
                        "product": "87OCT 10%ETH - SUNOCO",
                        "supply_owner_id": "6908f1eab4fcf6a36e6660e1"
                    },
                    {
                        "location_name": "1083",
                        "product": "93OCT 10%ETH - SUNOCO",
                        "supply_owner_id": "6908f1eab4fcf6a36e6660e1"
                    }
                ]
            },
            "freight_customer_id": "6908f1eab4fcf6a36e6660e1",
            "delivered": False,
            "out_of_network": False,
            "estimated_freight": {
                "id": "6953f4eda8bb5f01056eaca4",
                "parent": None,
                "created_at": "2025-12-30T15:51:09.909000",
                "type": "freight",
                "status": "open",
                "reversed": False,
                "transactions": [
                    {
                        "model_type": "base",
                        "type": "Base Freight",
                        "subtype": "Band 31-35",
                        "rate": 0.0341,
                        "amount": 4000,
                        "total": 136.4,
                        "uom": "gallons",
                        "errors": [],
                        "id": "6953f4eda8bb5f01056eaca1",
                        "product_group": "Gasoline",
                        "product_id": "6908f4469257414108faf071",
                        "product_name": "87OCT 10%ETH - SUNOCO",
                        "origin": "MARATHON -Detroit",
                        "origin_id": "69090ecad6ff2d863efaf07b",
                        "origin_override": None,
                        "origin_id_override": None,
                        "destination": "1082",
                        "destination_id": "69090f2e5a7ba1bea579a967",
                        "destination_override": None,
                        "destination_id_override": None,
                        "legs": [
                            {
                                "origin": "1082",
                                "origin_id": "69090f2e5a7ba1bea579a965",
                                "destination": "1083",
                                "destination_id": "69090f2e5a7ba1bea579a969",
                                "distance": 16.05
                            },
                            {
                                "origin": "MARATHON -Detroit",
                                "origin_id": "69090ecad6ff2d863efaf07b",
                                "destination": "1082",
                                "destination_id": "69090f2e5a7ba1bea579a965",
                                "distance": 14.37
                            }
                        ],
                        "use_surcharge": True,
                        "manual": False,
                        "has_dependency": False,
                        "created_date": None,
                        "requires_approval": False,
                        "requires_approval_reason_code": False,
                        "is_approved": False,
                        "approved_by": None,
                        "approved_date": None,
                        "approved_reason_code": None,
                        "amount_override": None,
                        "rate_override": None,
                        "product_id_override": None,
                        "product_name_override": None,
                        "product_group_override": None,
                        "subtype_override": None,
                        "uom_override": None,
                        "total_override": None,
                        "override_by": None,
                        "override_date": None,
                        "exclude_from_invoice": False,
                        "credit_rebill_metadata": None,
                        "extra_data": {
                            "mileage": 30
                        },
                        "gross_volume": None,
                        "net_volume": None,
                        "bol_date": None,
                        "bol_number": None,
                        "bol_number_override": None,
                        "delivery_date": None
                    },
                    {
                        "model_type": "base",
                        "type": "Base Freight",
                        "subtype": "Band 31-35",
                        "rate": 0.0341,
                        "amount": 6000,
                        "total": 204.6,
                        "uom": "gallons",
                        "errors": [],
                        "id": "6953f4eda8bb5f01056eaca2",
                        "product_group": "Gasoline",
                        "product_id": "6908f4469257414108faf071",
                        "product_name": "87OCT 10%ETH - SUNOCO",
                        "origin": "MARATHON -Detroit",
                        "origin_id": "69090ecad6ff2d863efaf07b",
                        "origin_override": None,
                        "origin_id_override": None,
                        "destination": "1083",
                        "destination_id": "69090f2e5a7ba1bea579a96b",
                        "destination_override": None,
                        "destination_id_override": None,
                        "legs": [
                            {
                                "origin": "1082",
                                "origin_id": "69090f2e5a7ba1bea579a965",
                                "destination": "1083",
                                "destination_id": "69090f2e5a7ba1bea579a969",
                                "distance": 16.05
                            },
                            {
                                "origin": "MARATHON -Detroit",
                                "origin_id": "69090ecad6ff2d863efaf07b",
                                "destination": "1082",
                                "destination_id": "69090f2e5a7ba1bea579a965",
                                "distance": 14.37
                            }
                        ],
                        "use_surcharge": True,
                        "manual": False,
                        "has_dependency": False,
                        "created_date": None,
                        "requires_approval": False,
                        "requires_approval_reason_code": False,
                        "is_approved": False,
                        "approved_by": None,
                        "approved_date": None,
                        "approved_reason_code": None,
                        "amount_override": None,
                        "rate_override": None,
                        "product_id_override": None,
                        "product_name_override": None,
                        "product_group_override": None,
                        "subtype_override": None,
                        "uom_override": None,
                        "total_override": None,
                        "override_by": None,
                        "override_date": None,
                        "exclude_from_invoice": False,
                        "credit_rebill_metadata": None,
                        "extra_data": {
                            "mileage": 30
                        },
                        "gross_volume": None,
                        "net_volume": None,
                        "bol_date": None,
                        "bol_number": None,
                        "bol_number_override": None,
                        "delivery_date": None
                    },
                    {
                        "model_type": "base",
                        "type": "Base Freight",
                        "subtype": "Band 31-35",
                        "rate": 0.0341,
                        "amount": 1000,
                        "total": 34.1,
                        "uom": "gallons",
                        "errors": [],
                        "id": "6953f4eda8bb5f01056eaca3",
                        "product_group": "Gasoline",
                        "product_id": "6908f409b4fcf6a36e6660f3",
                        "product_name": "93OCT 10%ETH - SUNOCO",
                        "origin": "MARATHON -Detroit",
                        "origin_id": "69090ecad6ff2d863efaf07b",
                        "origin_override": None,
                        "origin_id_override": None,
                        "destination": "1083",
                        "destination_id": "69090f2e5a7ba1bea579a96b",
                        "destination_override": None,
                        "destination_id_override": None,
                        "legs": [
                            {
                                "origin": "1082",
                                "origin_id": "69090f2e5a7ba1bea579a965",
                                "destination": "1083",
                                "destination_id": "69090f2e5a7ba1bea579a969",
                                "distance": 16.05
                            },
                            {
                                "origin": "MARATHON -Detroit",
                                "origin_id": "69090ecad6ff2d863efaf07b",
                                "destination": "1082",
                                "destination_id": "69090f2e5a7ba1bea579a965",
                                "distance": 14.37
                            }
                        ],
                        "use_surcharge": True,
                        "manual": False,
                        "has_dependency": False,
                        "created_date": None,
                        "requires_approval": False,
                        "requires_approval_reason_code": False,
                        "is_approved": False,
                        "approved_by": None,
                        "approved_date": None,
                        "approved_reason_code": None,
                        "amount_override": None,
                        "rate_override": None,
                        "product_id_override": None,
                        "product_name_override": None,
                        "product_group_override": None,
                        "subtype_override": None,
                        "uom_override": None,
                        "total_override": None,
                        "override_by": None,
                        "override_date": None,
                        "exclude_from_invoice": False,
                        "credit_rebill_metadata": None,
                        "extra_data": {
                            "mileage": 30
                        },
                        "gross_volume": None,
                        "net_volume": None,
                        "bol_date": None,
                        "bol_number": None,
                        "bol_number_override": None,
                        "delivery_date": None
                    }
                ],
                "counterparty_id": "6908f1eab4fcf6a36e6660e1",
                "book_type": "Cost",
                "effective_date_used": "2025-12-30T00:00:00",
                "invoice_group_id": None,
                "invoice_number": None,
                "accessorial_invoice_number": None,
                "new_version_available": False,
                "new_version_total": None,
                "error": None
            },
            "legs": [
                {
                    "origin": "1082",
                    "origin_id": "69090f2e5a7ba1bea579a965",
                    "destination": "1083",
                    "destination_id": "69090f2e5a7ba1bea579a969",
                    "distance": 16.05
                },
                {
                    "origin": "MARATHON -Detroit",
                    "origin_id": "69090ecad6ff2d863efaf07b",
                    "destination": "1082",
                    "destination_id": "69090f2e5a7ba1bea579a965",
                    "distance": 14.37
                }
            ],
            "freight_distance": 30.42,
            "supply_owner_id": "6908f1eab4fcf6a36e6660e1"
        },
        "created_by": "JBerninger",
        "note": None,
        "estimated_load_minutes": 44,
        "total_miles": 35.91836175356,
        "loaded_miles": 30.42233356032,
        "unloaded_miles": 5.49602819324,
        "reference_order_number": None,
        "extra_data": {},
        "tractor": "202",
        "trailer": "1609",
        "po_number": None
    },
    {
        "type": "regular",
        "order_id": "69581c65ee5aaa626ed7ddee",
        "drivers": [],
        "order_number": 1196,
        "order_date": "2026-01-02T19:28:37Z",
        "order_state": "complete",
        "carrier_window_start": None,
        "carrier_window_end": None,
        "carrier_window_start_local": None,
        "carrier_window_end_local": None,
        "carrier_notify_state": "none",
        "load_window_start": "2025-12-31T23:00:00Z",
        "load_window_end": "2026-01-06T23:00:00Z",
        "dispatch_window_start": "2026-01-03T06:00:00Z",
        "dispatch_window_end": "2026-01-03T12:00:00Z",
        "hauler_counterparty_id": "None",
        "hauler_counterparty_name": "",
        "hauler_counterparty_source_id": None,
        "hauler_counterparty_source_system": "",
        "hauled_by_updated_by": None,
        "hauled_by_updated": None,
        "loads": [
            {
                "location_id": "69090ecad6ff2d863efaf084",
                "location_name": "SUNOCO -RIVER ROUGE",
                "location_source_id": "T-38-MI-3030",
                "location_source_system": "PDI",
                "location_type": "Terminal",
                "details": [
                    {
                        "detail_id": "69581c65ee5aaa626ed7dde4",
                        "counterparty_id": "6908f1eab4fcf6a36e6660d9",
                        "counterparty_name": "Sunoco",
                        "counterparty_source_id": "5008",
                        "counterparty_source_system": None,
                        "price_type": "rack",
                        "contract": "",
                        "product_id": "6908f4469257414108faf071",
                        "product_name": "87OCT 10%ETH - SUNOCO",
                        "product_source_id": "FO-87-E10 | SUN",
                        "product_source_system": "PDI",
                        "drop_product_id": "6908f4469257414108faf071",
                        "drop_product_name": "87OCT 10%ETH - SUNOCO",
                        "drop_product_source_id": "FO-87-E10 | SUN",
                        "drop_product_source_system": "PDI",
                        "quantity": 4000,
                        "compartment_index": 0,
                        "blend_code": None,
                        "product_codes": [],
                        "load_number": None
                    },
                    {
                        "detail_id": "69581c65ee5aaa626ed7dde4",
                        "counterparty_id": "6908f1eab4fcf6a36e6660d9",
                        "counterparty_name": "Sunoco",
                        "counterparty_source_id": "5008",
                        "counterparty_source_system": None,
                        "price_type": "rack",
                        "contract": "",
                        "product_id": "6908f4469257414108faf071",
                        "product_name": "87OCT 10%ETH - SUNOCO",
                        "product_source_id": "FO-87-E10 | SUN",
                        "product_source_system": "PDI",
                        "drop_product_id": "6908f4469257414108faf071",
                        "drop_product_name": "87OCT 10%ETH - SUNOCO",
                        "drop_product_source_id": "FO-87-E10 | SUN",
                        "drop_product_source_system": "PDI",
                        "quantity": 6000,
                        "compartment_index": 1,
                        "blend_code": None,
                        "product_codes": [],
                        "load_number": None
                    },
                    {
                        "detail_id": "69581c65ee5aaa626ed7dde5",
                        "counterparty_id": "6908f1eab4fcf6a36e6660d9",
                        "counterparty_name": "Sunoco",
                        "counterparty_source_id": "5008",
                        "counterparty_source_system": None,
                        "price_type": "rack",
                        "contract": "",
                        "product_id": "6908f409b4fcf6a36e6660f3",
                        "product_name": "93OCT 10%ETH - SUNOCO",
                        "product_source_id": "FO-93-E10 | SUN",
                        "product_source_system": "PDI",
                        "drop_product_id": "6908f409b4fcf6a36e6660f3",
                        "drop_product_name": "93OCT 10%ETH - SUNOCO",
                        "drop_product_source_id": "FO-93-E10 | SUN",
                        "drop_product_source_system": "PDI",
                        "quantity": 1000,
                        "compartment_index": 2,
                        "blend_code": None,
                        "product_codes": [],
                        "load_number": None
                    }
                ],
                "eta": None,
                "status": "unset"
            }
        ],
        "drops": [
            {
                "location_id": "69090f2e5a7ba1bea579a965",
                "location_name": "1082",
                "location_source_id": None,
                "location_source_system": "",
                "location_type": "Store",
                "details": [
                    {
                        "detail_id": "69581c65ee5aaa626ed7dde6",
                        "counterparty_id": "",
                        "counterparty_name": "",
                        "counterparty_source_id": None,
                        "counterparty_source_system": None,
                        "product_id": "6908f4469257414108faf071",
                        "product_name": "87OCT 10%ETH - SUNOCO",
                        "product_source_id": "FO-87-E10 | SUN",
                        "product_source_system": "PDI",
                        "quantity": 4000,
                        "tank_id": 1,
                        "sources": [
                            {
                                "compartment_index": 0,
                                "volume": 4000
                            }
                        ],
                        "destination_codes": [],
                        "blend_codes": []
                    }
                ],
                "eta": None,
                "status": "unset"
            },
            {
                "location_id": "69090f2e5a7ba1bea579a969",
                "location_name": "1083",
                "location_source_id": None,
                "location_source_system": "",
                "location_type": "Store",
                "details": [
                    {
                        "detail_id": "69581c65ee5aaa626ed7dde7",
                        "counterparty_id": "",
                        "counterparty_name": "",
                        "counterparty_source_id": None,
                        "counterparty_source_system": None,
                        "product_id": "6908f4469257414108faf071",
                        "product_name": "87OCT 10%ETH - SUNOCO",
                        "product_source_id": "FO-87-E10 | SUN",
                        "product_source_system": "PDI",
                        "quantity": 6000,
                        "tank_id": 3,
                        "sources": [
                            {
                                "compartment_index": 1,
                                "volume": 6000
                            }
                        ],
                        "destination_codes": [],
                        "blend_codes": []
                    },
                    {
                        "detail_id": "69581c65ee5aaa626ed7dde8",
                        "counterparty_id": "",
                        "counterparty_name": "",
                        "counterparty_source_id": None,
                        "counterparty_source_system": None,
                        "product_id": "6908f409b4fcf6a36e6660f3",
                        "product_name": "93OCT 10%ETH - SUNOCO",
                        "product_source_id": "FO-93-E10 | SUN",
                        "product_source_system": "PDI",
                        "quantity": 1000,
                        "tank_id": 1,
                        "sources": [
                            {
                                "compartment_index": 2,
                                "volume": 1000
                            }
                        ],
                        "destination_codes": [],
                        "blend_codes": []
                    }
                ],
                "eta": None,
                "status": "unset"
            }
        ],
        "detours": [],
        "trip_status": "complete",
        "last_change_date": "2026-01-02T19:29:59.497000",
        "market": "Michigan",
        "supply_option": {
            "option": "recommended",
            "carrier_id": None,
            "carrier": None,
            "loaded_miles": 29.655561509392,
            "freight": 0.0324,
            "freight_cost": 0,
            "stops": 3,
            "prices": [
                {
                    "product_name": "87OCT 10%ETH - SUNOCO",
                    "volume": 10000,
                    "components": [
                        {
                            "timezone_data": None,
                            "curve_id": "6945cec2f88ce49ef5fc509c",
                            "price_id": "6955ab1fb816b69ba5c18eca",
                            "terminal_id": "69090ecad6ff2d863efaf084",
                            "terminal": "SUNOCO -RIVER ROUGE",
                            "supplier_id": "6908f1eab4fcf6a36e6660d9",
                            "supplier": "Sunoco",
                            "price_type": "rack",
                            "product_id": "6908f4469257414108faf071",
                            "product": "87OCT 10%ETH - SUNOCO",
                            "supply_owner_id": "6908f1eab4fcf6a36e6660e1",
                            "supply_owner": "Bazco",
                            "price": 1.7556,
                            "contract": "",
                            "volume": 10000,
                            "effective_from": "2025-12-31T23:00:00Z",
                            "effective_to": "2026-01-06T23:00:00Z",
                            "group_id": None,
                            "group_effective_cutover": None,
                            "group_effective_cutover_timezone": None,
                            "group_identifier": None,
                            "group_name": None,
                            "min_constraint": None,
                            "max_constraint": None,
                            "load_number": None,
                            "load_numbers": [],
                            "directive_info": None
                        }
                    ],
                    "product_total": 17556,
                    "blend_code": None
                },
                {
                    "product_name": "93OCT 10%ETH - SUNOCO",
                    "volume": 1000,
                    "components": [
                        {
                            "timezone_data": None,
                            "curve_id": "6945cec2f88ce49ef5fc505a",
                            "price_id": "6955ab1fb816b69ba5c18ea9",
                            "terminal_id": "69090ecad6ff2d863efaf084",
                            "terminal": "SUNOCO -RIVER ROUGE",
                            "supplier_id": "6908f1eab4fcf6a36e6660d9",
                            "supplier": "Sunoco",
                            "price_type": "rack",
                            "product_id": "6908f409b4fcf6a36e6660f3",
                            "product": "93OCT 10%ETH - SUNOCO",
                            "supply_owner_id": "6908f1eab4fcf6a36e6660e1",
                            "supply_owner": "Bazco",
                            "price": 2.8061,
                            "contract": "",
                            "volume": 1000,
                            "effective_from": "2025-12-31T23:00:00Z",
                            "effective_to": "2026-01-06T23:00:00Z",
                            "group_id": None,
                            "group_effective_cutover": None,
                            "group_effective_cutover_timezone": None,
                            "group_identifier": None,
                            "group_name": None,
                            "min_constraint": None,
                            "max_constraint": None,
                            "load_number": None,
                            "load_numbers": [],
                            "directive_info": None
                        }
                    ],
                    "product_total": 2806.1,
                    "blend_code": None
                }
            ],
            "product_total": 20362.1,
            "landed_cost": 20718.5,
            "supply_owner_map": {
                "definitions": [
                    {
                        "location_name": "1082",
                        "product": "87OCT 10%ETH - SUNOCO",
                        "supply_owner_id": "6908f1eab4fcf6a36e6660e1"
                    },
                    {
                        "location_name": "1083",
                        "product": "87OCT 10%ETH - SUNOCO",
                        "supply_owner_id": "6908f1eab4fcf6a36e6660e1"
                    },
                    {
                        "location_name": "1083",
                        "product": "93OCT 10%ETH - SUNOCO",
                        "supply_owner_id": "6908f1eab4fcf6a36e6660e1"
                    }
                ]
            },
            "freight_customer_id": "6908f1eab4fcf6a36e6660e1",
            "delivered": False,
            "out_of_network": False,
            "estimated_freight": {
                "id": "69581c65ee5aaa626ed7ddec",
                "parent": None,
                "created_at": "2026-01-02T19:28:37.447000",
                "type": "freight",
                "status": "open",
                "reversed": False,
                "transactions": [
                    {
                        "model_type": "base",
                        "type": "Base Freight",
                        "subtype": "Band 26-30",
                        "rate": 0.0324,
                        "amount": 4000,
                        "total": 129.6,
                        "uom": "gallons",
                        "errors": [],
                        "id": "69581c65ee5aaa626ed7dde9",
                        "product_group": "Gasoline",
                        "product_id": "6908f4469257414108faf071",
                        "product_name": "87OCT 10%ETH - SUNOCO",
                        "origin": "SUNOCO -RIVER ROUGE",
                        "origin_id": "69090ecad6ff2d863efaf084",
                        "origin_override": None,
                        "origin_id_override": None,
                        "destination": "1082",
                        "destination_id": "69090f2e5a7ba1bea579a967",
                        "destination_override": None,
                        "destination_id_override": None,
                        "legs": [
                            {
                                "origin": "SUNOCO -RIVER ROUGE",
                                "origin_id": "69090ecad6ff2d863efaf084",
                                "destination": "1082",
                                "destination_id": "69090f2e5a7ba1bea579a965",
                                "distance": 13.6
                            },
                            {
                                "origin": "1082",
                                "origin_id": "69090f2e5a7ba1bea579a965",
                                "destination": "1083",
                                "destination_id": "69090f2e5a7ba1bea579a969",
                                "distance": 16.05
                            }
                        ],
                        "use_surcharge": True,
                        "manual": False,
                        "has_dependency": False,
                        "created_date": None,
                        "requires_approval": False,
                        "requires_approval_reason_code": False,
                        "is_approved": False,
                        "approved_by": None,
                        "approved_date": None,
                        "approved_reason_code": None,
                        "amount_override": None,
                        "rate_override": None,
                        "product_id_override": None,
                        "product_name_override": None,
                        "product_group_override": None,
                        "subtype_override": None,
                        "uom_override": None,
                        "total_override": None,
                        "override_by": None,
                        "override_date": None,
                        "exclude_from_invoice": False,
                        "credit_rebill_metadata": None,
                        "extra_data": {
                            "mileage": 30
                        },
                        "gross_volume": None,
                        "net_volume": None,
                        "bol_date": None,
                        "bol_number": None,
                        "bol_number_override": None,
                        "delivery_date": None
                    },
                    {
                        "model_type": "base",
                        "type": "Base Freight",
                        "subtype": "Band 26-30",
                        "rate": 0.0324,
                        "amount": 6000,
                        "total": 194.4,
                        "uom": "gallons",
                        "errors": [],
                        "id": "69581c65ee5aaa626ed7ddea",
                        "product_group": "Gasoline",
                        "product_id": "6908f4469257414108faf071",
                        "product_name": "87OCT 10%ETH - SUNOCO",
                        "origin": "SUNOCO -RIVER ROUGE",
                        "origin_id": "69090ecad6ff2d863efaf084",
                        "origin_override": None,
                        "origin_id_override": None,
                        "destination": "1083",
                        "destination_id": "69090f2e5a7ba1bea579a96b",
                        "destination_override": None,
                        "destination_id_override": None,
                        "legs": [
                            {
                                "origin": "SUNOCO -RIVER ROUGE",
                                "origin_id": "69090ecad6ff2d863efaf084",
                                "destination": "1082",
                                "destination_id": "69090f2e5a7ba1bea579a965",
                                "distance": 13.6
                            },
                            {
                                "origin": "1082",
                                "origin_id": "69090f2e5a7ba1bea579a965",
                                "destination": "1083",
                                "destination_id": "69090f2e5a7ba1bea579a969",
                                "distance": 16.05
                            }
                        ],
                        "use_surcharge": True,
                        "manual": False,
                        "has_dependency": False,
                        "created_date": None,
                        "requires_approval": False,
                        "requires_approval_reason_code": False,
                        "is_approved": False,
                        "approved_by": None,
                        "approved_date": None,
                        "approved_reason_code": None,
                        "amount_override": None,
                        "rate_override": None,
                        "product_id_override": None,
                        "product_name_override": None,
                        "product_group_override": None,
                        "subtype_override": None,
                        "uom_override": None,
                        "total_override": None,
                        "override_by": None,
                        "override_date": None,
                        "exclude_from_invoice": False,
                        "credit_rebill_metadata": None,
                        "extra_data": {
                            "mileage": 30
                        },
                        "gross_volume": None,
                        "net_volume": None,
                        "bol_date": None,
                        "bol_number": None,
                        "bol_number_override": None,
                        "delivery_date": None
                    },
                    {
                        "model_type": "base",
                        "type": "Base Freight",
                        "subtype": "Band 26-30",
                        "rate": 0.0324,
                        "amount": 1000,
                        "total": 32.4,
                        "uom": "gallons",
                        "errors": [],
                        "id": "69581c65ee5aaa626ed7ddeb",
                        "product_group": "Gasoline",
                        "product_id": "6908f409b4fcf6a36e6660f3",
                        "product_name": "93OCT 10%ETH - SUNOCO",
                        "origin": "SUNOCO -RIVER ROUGE",
                        "origin_id": "69090ecad6ff2d863efaf084",
                        "origin_override": None,
                        "origin_id_override": None,
                        "destination": "1083",
                        "destination_id": "69090f2e5a7ba1bea579a96b",
                        "destination_override": None,
                        "destination_id_override": None,
                        "legs": [
                            {
                                "origin": "SUNOCO -RIVER ROUGE",
                                "origin_id": "69090ecad6ff2d863efaf084",
                                "destination": "1082",
                                "destination_id": "69090f2e5a7ba1bea579a965",
                                "distance": 13.6
                            },
                            {
                                "origin": "1082",
                                "origin_id": "69090f2e5a7ba1bea579a965",
                                "destination": "1083",
                                "destination_id": "69090f2e5a7ba1bea579a969",
                                "distance": 16.05
                            }
                        ],
                        "use_surcharge": True,
                        "manual": False,
                        "has_dependency": False,
                        "created_date": None,
                        "requires_approval": False,
                        "requires_approval_reason_code": False,
                        "is_approved": False,
                        "approved_by": None,
                        "approved_date": None,
                        "approved_reason_code": None,
                        "amount_override": None,
                        "rate_override": None,
                        "product_id_override": None,
                        "product_name_override": None,
                        "product_group_override": None,
                        "subtype_override": None,
                        "uom_override": None,
                        "total_override": None,
                        "override_by": None,
                        "override_date": None,
                        "exclude_from_invoice": False,
                        "credit_rebill_metadata": None,
                        "extra_data": {
                            "mileage": 30
                        },
                        "gross_volume": None,
                        "net_volume": None,
                        "bol_date": None,
                        "bol_number": None,
                        "bol_number_override": None,
                        "delivery_date": None
                    }
                ],
                "counterparty_id": "6908f1eab4fcf6a36e6660e1",
                "book_type": "Cost",
                "effective_date_used": "2026-01-03T00:00:00",
                "invoice_group_id": None,
                "invoice_number": None,
                "accessorial_invoice_number": None,
                "new_version_available": False,
                "new_version_total": None,
                "error": None
            },
            "legs": [
                {
                    "origin": "SUNOCO -RIVER ROUGE",
                    "origin_id": "69090ecad6ff2d863efaf084",
                    "destination": "1082",
                    "destination_id": "69090f2e5a7ba1bea579a965",
                    "distance": 13.6
                },
                {
                    "origin": "1082",
                    "origin_id": "69090f2e5a7ba1bea579a965",
                    "destination": "1083",
                    "destination_id": "69090f2e5a7ba1bea579a969",
                    "distance": 16.05
                }
            ],
            "freight_distance": 29.65,
            "supply_owner_id": "6908f1eab4fcf6a36e6660e1"
        },
        "created_by": "felipe.bucheli",
        "note": None,
        "estimated_load_minutes": 43,
        "total_miles": 29.655561509392,
        "loaded_miles": 29.655561509392,
        "unloaded_miles": 0,
        "reference_order_number": None,
        "extra_data": {},
        "tractor": None,
        "trailer": None,
        "po_number": None
    }
]


def create_mock_sd_client() -> GravitateSDAPI:
    mock = MagicMock(spec=GravitateSDAPI)
    mock.get_bols_and_drops = AsyncMock(return_value=_mock_response(mock_bols_and_drops_response))
    mock.get_orders = AsyncMock(return_value=_mock_response(mock_get_orders_response))
    return mock


def _mock_response(json_data) -> MagicMock:
    """Create a mock httpx Response."""
    response = MagicMock(spec=Response)
    response.status_code = 200
    response.json.return_value = json_data
    response.raise_for_status = MagicMock()
    return response


if __name__ == "__main__":
    @hatch_eggs
    async def main(sd_client:GravitateSDAPI = Egg(create_mock_sd_client)):
        results = await AllocationMatcher.match_order_numbers([1103, 1196], sd_client)
        print(f"Matched {len(results)} allocations:\n")
        for r in results:
            print(f'Order Number: {r.order_number}')
            print(f"Store {r.allocated_bol.store_number}, Tank {r.allocated_bol.store_tank}")
            print(f"  Product: {r.allocated_bol.store_product}")
            print(f"  Planned: {r.planned_quantity} gal")
            print(f"  Actual:  {r.actual_quantity} gal")
            print(f"  Variance: {r.variance:+d} gal ({r.variance_pct:.2f}%)")
            print()


    asyncio.run(main())



import json
from math import ceil
from time import sleep
from typing import Dict, List

from loguru import logger
from more_itertools import chunked

from bb_integrations_lib.gravitate.sd_api import GravitateSDAPI
from bb_integrations_lib.models.pipeline_structs import BBDUploadResult
from bb_integrations_lib.models.rita.issue import IssueBase, IssueCategory
from bb_integrations_lib.protocols.flat_file import TankSales
from bb_integrations_lib.protocols.pipelines import Step
from bb_integrations_lib.util.utils import CustomJSONEncoder


class BBDUploadTankSalesStep(Step):
    """
    Takes a list of TankSales and uploads them to Best Buy
    """

    def __init__(self, sd_client: GravitateSDAPI, sleep_between: float = 0.5, chunk_size: int = 1000, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sd_client = sd_client
        self.sleep_between = sleep_between
        self.chunk_size = chunk_size

    def describe(self) -> str:
        return "Upload Tanksales to BBD"

    async def execute(self, i: List[TankSales]) -> BBDUploadResult:
        logs = {"requests": [], "responses": [], "errors": []}
        try:
            total_sales = len(i)
            count = ceil(total_sales / self.chunk_size)
            attempted = 0
            succeeded = 0
            failed_items = []
            store_ids = []

            for idx, group in enumerate(chunked(i, self.chunk_size)):
                logger.info(f"Uploading sales to bestbuy {idx + 1} of {count} to: {self.sd_client.base_url}  ")
                attempted += len(group)
                serialized_group = [g.model_dump(mode="json") for g in group]
                batch_store_ids = [g.get("store_number", "unknown") for g in serialized_group]
                logs["requests"].append({
                    "row_id": idx,
                    "request": serialized_group,
                    "store_ids": batch_store_ids
                })

                try:
                    response: Dict = await self.sd_client.upload_tank_sales(serialized_group)
                    response_data = response
                    logs["responses"].append({
                        "row_id": idx,
                        "response": response_data,
                    })
                    created = response_data.get("created", 0)
                    updated = response_data.get("updated", 0)
                    failed = response_data.get("failed", [])
                    current_succeeded = created + updated
                    succeeded += current_succeeded
                    if current_succeeded > 0:
                        if current_succeeded == len(serialized_group) and not failed:
                            store_ids.extend(batch_store_ids)
                        else:
                            failed_store_numbers = [f["record"]["store_number"] for f in failed if
                                                    "record" in f and "store_number" in f["record"]]
                            failed_set = set(failed_store_numbers)
                            successful_ids = [
                                store_id for store_id in batch_store_ids
                                if store_id not in failed_set
                            ]
                            store_ids.extend(successful_ids)
                    if failed:
                        failed_items.extend(failed)
                        logs["errors"].append({
                            "row_id": idx,
                            "failed_items": failed,
                            "response": response_data
                        })
                        logger.error(f"Errors occurred while uploading data: {failed}")
                    logger.info(f"Batch {idx + 1}: Created {created}, Updated {updated}, Failed {len(failed)}")
                    sleep(self.sleep_between)
                except Exception as e:
                    error_msg = f"Batch {idx} sales failed | {e}"
                    logger.error(error_msg)
                    failed_items.extend(batch_store_ids)
                    logs["errors"].append({
                        "row_id": idx,
                        "exception": str(e),
                        "store_ids": batch_store_ids
                    })
                    continue

            logger.info(f"Successfully uploaded {succeeded} of {attempted} sales.")
            logger.info(f"Failed to upload {len(failed_items)} of {attempted} sales")
            if failed_items and hasattr(self.pipeline_context,
                                        'issue_report_config') and self.pipeline_context.issue_report_config:
                irc = self.pipeline_context.issue_report_config
                fc = self.pipeline_context.file_config
                key = f"{irc.key_base}_{fc.config_id}_failed_to_upload"
                self.pipeline_context.issues.append(IssueBase(
                    key=key,
                    config_id=fc.config_id,
                    name="Failed to upload Tanksales",
                    category=IssueCategory.TANK_READING,
                    problem_short=f"{len(failed_items)} sales did not upload",
                    problem_long=json.dumps(failed_items)
                ))

            self.pipeline_context.included_files["sales data upload"] = json.dumps(logs, cls=CustomJSONEncoder)
            return BBDUploadResult(
                succeeded=succeeded,
                failed=len(failed_items),
                succeeded_items=store_ids
            )

        except Exception as e:
            logger.exception(f"Unable to upload | {e}")
            raise e



import asyncio
from datetime import datetime, UTC, timedelta
from typing import Any, Optional, Literal
from zoneinfo import ZoneInfo

from bson import ObjectId
from httpx import HTTPStatusError
from loguru import logger
from pydantic import BaseModel, ValidationError
from pymongo import MongoClient
from pymongo.asynchronous.database import AsyncDatabase

from bb_integrations_lib.gravitate.sd_api import GravitateSDAPI
from bb_integrations_lib.protocols.pipelines import Step
from bb_integrations_lib.provider.api.platform_science.client import PlatformScienceClient
from bb_integrations_lib.provider.api.platform_science.model import JobDefinition, ShipmentDetails, ValueWithUnit, \
    JobLocation, JobStep, JobTask, LoadDefinition, LoadEntity
from bb_integrations_lib.util.utils import lookup


class PlatSciLink(BaseModel):
    """Platform Science linkage details to be stored in S&D order_v2 extra_data."""
    job_id: str
    completed: bool
    last_order_state: str


class PlatformScienceOrderSyncStep(Step):
    """
    Export the current status of a Gravitate order to Platform Science workflow, either creating, updating, or
    completing as necessary.
    """

    def __init__(
            self,
            sd_client: GravitateSDAPI,
            psc: PlatformScienceClient,
            mongo_database: AsyncDatabase,
            order_nums: list[int] | None = None,
            ps_link_key: str = "platform_science",
            timezone: ZoneInfo = UTC,
            *args, **kwargs
    ):
        super().__init__(*args, *kwargs)
        self.sd_client = sd_client
        self.psc = psc
        self.db = mongo_database

        self.order_nums = order_nums

        self.loc_lkp: Optional[dict] = None
        self.dt_lkp: Optional[dict] = None
        self.ps_link_key = ps_link_key
        self.timezone = timezone

    def describe(self) -> str:
        return "Update Platform Science workflow from Gravitate order status"

    async def execute(self, i: Any) -> None:
        if self.loc_lkp is None:
            logger.info("Fetching locations from S&D")
            locations = await self.sd_client.all_locations()
            self.loc_lkp = lookup(locations.json(), lambda x: x["id"])

        if self.order_nums:
            logger.warning(f"Using provided order numbers for testing: {self.order_nums}")
            changed_orders = []
            for i, order_num in enumerate(self.order_nums):
                logger.info(f"Downloading order {order_num} ({i+1}/{len(self.order_nums)})")
                order_resp = await self.sd_client.get_orders(order_number=order_num)
                order_resp.raise_for_status()
                changed_orders.extend(order_resp.json())
        else:
            change_window = 60
            logger.info(f"Fetching orders changed in the last {change_window} minutes from S&D")
            changed_orders_resp = await self.sd_client.get_orders(
                last_change_date=datetime.now(tz=UTC) - timedelta(minutes=change_window)
            )
            changed_orders = changed_orders_resp.json()

        if not changed_orders:
            logger.info(f"No changed orders, stopping")
            return

        if self.dt_lkp is None:
            self.dt_lkp = {}
            driver_tracking = await self.sd_client.get_driver_tracking(
                order_numbers=[x["order_number"] for x in changed_orders])
            for dt_item in driver_tracking.json():
                for ao in dt_item["assigned_orders"]:
                    self.dt_lkp[ao["number"]] = dt_item

        logger.info("Ordering orders by status: completions, updates, creations")
        completions, updates, creations = self._categorize_orders(changed_orders)
        logger.info(f"Processing {len(completions)} completions, {len(updates)} updates, {len(creations)} creations")
        ordered_orders = completions + updates + creations

        for order in ordered_orders:
            try:
                with logger.contextualize(order_number=order["order_number"]):
                    logger.info(f"Syncing order {order['order_number']}")
                    if order["order_number"] in self.dt_lkp.keys():
                        await self.sync_to_platform_science(order)
                    else:
                        logger.warning("No driver tracking record found for this order, skipping")
            except HTTPStatusError as he:
                if he.response.status_code == 404:
                    logger.warning("Got a 404 from PS syncing the order, resetting PS link.")
                    self._clear_ps_link(order["order_id"])
                    # Also have to clear out the extra_data field in the cached order object
                    order.get("extra_data", {}).pop(self.ps_link_key, None)
                    logger.info("Recreating PS order")
                    try:
                        await self.sync_to_platform_science(order)
                    except HTTPStatusError as retry_he:
                        logger.exception(f"Failed to recreate PS order after 404: {retry_he}: {retry_he.response.text}")
                    except Exception as retry_e:
                        logger.exception(f"Failed to recreate PS order after 404: {retry_e}")
                else:
                    logger.exception(f"Failed to sync order {order['order_number']}: {he}: {he.response.text}")
            except Exception as e:
                logger.exception(f"Failed to sync order {order['order_number']}: {e}")

    async def sync_to_platform_science(self, order: dict) -> None:
        # First, figure out if we have an existing PS workflow for this order
        # If we do, update the workflow. If not, create a new one. If we do have a workflow and this order is now
        # completed, attempt to complete the workflow.
        grav_order_state = order["order_state"]
        extra_data = order.get("extra_data", {})
        ps_link_raw = extra_data.get(self.ps_link_key)
        try:
            ps_link = PlatSciLink.model_validate(ps_link_raw)
        except ValidationError:
            if ps_link_raw is not None:
                logger.warning(f"Malformed Platform Science link data from extra_data: {ps_link_raw}")
            ps_link = None

        driver = self._get_driver(order)
        logger.info(f"For driver {driver}")

        # Do we have good link data?
        if ps_link:
            # Is it updatable?
            if not ps_link.completed:
                if grav_order_state != "complete":
                    logger.info(f"Updating an existing PS workflow ({ps_link.job_id})")
                    await self.update_existing_ps_workflow(driver, ps_link.job_id, order)
                    if grav_order_state == "in progress" and ps_link.last_order_state != "in progress":
                        logger.info("Order transitioned to in progress, creating load")
                        try:
                            await self.create_load(order, driver)
                        except HTTPStatusError as he:
                            logger.warning(f"{he.response.status_code}: {he.response.text}")
                else:
                    logger.info(f"Completing an uncompleted PS workflow ({ps_link.job_id})")
                    await self.complete_existing_ps_workflow(driver, ps_link.job_id, order)
            else:
                logger.info("PS workflow already completed, cannot send further updates")
        else:
            if driver:
                logger.info("No PS workflow found, creating new")
                await self.create_new_ps_workflow(order, driver)
                if grav_order_state == "in progress":
                    logger.info("New order is in progress, creating load")
                    await self.create_load(order, driver)
            else:
                logger.error("No driver on order - must have a driver to create in Platform Science")

    @staticmethod
    def order_is_pre_assign(order: dict) -> bool:
        return order["order_state"] == "accepted"

    async def create_load(self, order: dict, driver: str) -> None:
        entities = [
            LoadEntity(
                type="bill_of_lading",
                value=str(order["order_number"])
            ),
        ]
        if trailer := order.get("trailer"):
            entities.append(LoadEntity(
                type="trailer",
                value=str(trailer)
            ))

        start_date = datetime.now(self.timezone).date()
        # Get the latest drop ETA for this order, or set end_date = start_date if there are no drops / ETAs
        end_date = datetime.min
        for drop in order["drops"]:
            eta = drop.get("eta")
            if eta:
                eta = datetime.fromisoformat(eta)
                end_date = max(end_date, eta)
        if end_date == datetime.min:
            end_date = start_date
        else:
            end_date = end_date.astimezone(self.timezone).date()

        resp = await self.psc.create_load(
            driver,
            LoadDefinition(
                start_date=start_date,
                end_date=end_date,
                # TODO: Depends on this func getting called only when the order is completed
                user_external_id=driver,
                load=None,
                entities=entities
            )
        )
        if resp.is_error:
            logger.error(resp.json())
            resp.raise_for_status()

    async def create_new_ps_workflow(self, order: dict, driver: str) -> None:
        resp = await self.psc.create_workflow_job(
            driver,
            job_definition=self._convert_grav_order_to_job_definition(
                order, self.loc_lkp, self.dt_lkp, pre_assign=self.order_is_pre_assign(order)
            )
        )
        if resp.is_error:
            logger.error(resp.json())
            resp.raise_for_status()
        body = resp.json()
        logger.info(f"Created order, PS response {body}")
        logger.info("Setting S&D extra_data")
        job_id = str(body["data"]["job_id"])

        order_completed = order["order_state"] == "complete"
        if order_completed:
            logger.info("Order already completed, completing PS workflow")
            await self.complete_existing_ps_workflow(driver, job_id, order)
        else:
            await self._save_ps_link(
                order["order_id"],
                PlatSciLink(job_id=job_id, completed=order_completed, last_order_state=order["order_state"])
            )

    async def update_existing_ps_workflow(self, driver_id: str, job_id: str, order: dict) -> None:
        resp = await self.psc.update_workflow_job(
            driver_id,
            job_id,
            self._convert_grav_order_to_job_definition(order, self.loc_lkp, self.dt_lkp,
                                                       pre_assign=self.order_is_pre_assign(order))
        )
        resp.raise_for_status()
        await self._save_ps_link(
            order["order_id"],
            PlatSciLink(job_id=job_id, completed=False, last_order_state=order["order_state"])
        )

    async def complete_existing_ps_workflow(self, driver: str, ps_job_id: str, order: dict) -> None:
        # last_change_date is okay here since the likely last update on the order was when it was completed, if this job
        # runs soon after.
        await self.update_existing_ps_workflow(driver, ps_job_id, order)
        await self.psc.complete_workflow_job(ps_job_id, datetime.fromisoformat(order["last_change_date"]))
        await self._save_ps_link(
            order["order_id"],
            PlatSciLink(job_id=ps_job_id, completed=True, last_order_state=order["order_state"])
        )

    async def _save_ps_link(self, order_id: str, psl: PlatSciLink):
        ur = await self.db["order_v2"].update_one(
            filter={
                "_id": ObjectId(order_id)
            },
            update={
                "$set": {
                    f"extra_data.{self.ps_link_key}": psl.model_dump(mode="json")
                }
            }
        )
        logger.debug(f"Saved PS link: {psl}")

    async def _clear_ps_link(self, order_id: str):
        ur = await self.db["order_v2"].update_one(
            filter={
                "_id": ObjectId(order_id)
            },
            update={
                "$unset": {
                    f"extra_data.{self.ps_link_key}": ""
                }
            }
        )
        logger.debug(f"Cleared PS link for order_id '{order_id}'")

    @staticmethod
    def _get_driver(order: dict) -> str | None:
        drivers = order.get("drivers", [])
        if not drivers:
            logger.warning("No drivers found for order")
            return None
        if len(drivers) > 1:
            # Sync job uses this, so it should match the external_id field in Platform Science
            primary_driver = drivers[0]["username"]
            logger.warning(f"More than one driver found for order, using first driver ({primary_driver})")
        else:
            primary_driver = drivers[0]["username"]
        return primary_driver

    def _bulk_get_ps_links(self, order_ids: list[str]) -> dict[str, Optional[PlatSciLink]]:
        """Get PS links for a list of order ids"""
        results = self.db["order_v2"].find(
            filter={
                "_id": {
                    "$in": [ObjectId(oid) for oid in order_ids]
                }
            },
            projection={
                "_id": 1,
                f"extra_data.{self.ps_link_key}": 1
            }
        )

        ps_links = PlatformScienceOrderSyncStep._validate_ps_links_by_order_id(results, self.ps_link_key)
        return ps_links

    @staticmethod
    def _validate_ps_links_by_order_id(results: Cursor, ps_link_key: str) -> dict[str, Optional[PlatSciLink]]:
        ps_links = {}
        for doc in results:
            order_id = str(doc["_id"])
            ps_link_raw = doc.get("extra_data", {}).get(ps_link_key)
            try:
                ps_link = PlatSciLink.model_validate(ps_link_raw) if ps_link_raw else None
            except ValidationError:
                logger.warning(f"Bad Platform Science link data for order {order_id}: {ps_link_raw}")
                ps_link = None
            ps_links[order_id] = ps_link
        return ps_links

    def _categorize_orders(self, changed_orders: list[dict]) -> tuple[list[dict], list[dict], list[dict]]:
        """
        Categorize orders into completions, updates, and creations based on PS link status.
        """
        order_ids = [order["order_id"] for order in changed_orders]
        ps_links = self._bulk_get_ps_links(order_ids)

        completions = []
        updates = []
        creations = []

        for order in changed_orders:
            order_id = order["order_id"]
            ps_link = ps_links.get(order_id)

            if order["order_number"] not in self.dt_lkp.keys():
                logger.warning(
                    f"No driver tracking record found for order {order['order_number']}, skipping order...")
                continue

            if ps_link and not ps_link.completed:
                if order["order_state"] == "complete":
                    completions.append(order)
                else:
                    updates.append(order)
            else:
                creations.append(order)

        return completions, updates, creations

    @staticmethod
    def _convert_grav_order_to_job_definition(order: dict, loc_lkp: dict, dt_lkp: dict,
                                              pre_assign: bool) -> JobDefinition:
        loads = order["loads"]
        drops = order["drops"]
        return JobDefinition(
            status="pre_assign" if pre_assign else "active",
            external_id=str(order["order_number"]),
            locations=[PlatformScienceOrderSyncStep._convert_to_job_location(load, loc_lkp)
                       for load in loads] +
                      [PlatformScienceOrderSyncStep._convert_to_job_location(drop, loc_lkp)
                       for drop in drops],
            steps=PlatformScienceOrderSyncStep._loads_and_drops_to_steps(order["loads"], order["drops"],
                                                                         order["order_number"], dt_lkp),
            shipment_details=ShipmentDetails(
                total_distance=ValueWithUnit(
                    value=order["total_miles"],
                    uom="miles"
                )
            )
        )

    @staticmethod
    def _convert_to_job_location(load_or_drop: dict, loc_lkp: dict) -> JobLocation:
        """
        Loads and drops have a very similar (identical?) structure, so we can reuse this method for either case.
        Most of the data comes from the S&D location lookup anyway.
        """
        sd_loc = loc_lkp[load_or_drop["location_id"]]
        return JobLocation(
            external_id=load_or_drop["location_id"],  # TODO: Do we want to expose the location ID here?
            name=load_or_drop["location_name"],
            address=sd_loc["address"],
            latitude=f"{sd_loc['lat']:.4f}",
            longitude=f"{sd_loc['lon']:.4f}",
            city=sd_loc["city"],
            state=sd_loc["state"],
            country_code="US",  # TODO: Don't hardcode this
        )

    @staticmethod
    def _loads_and_drops_to_steps(loads: list[dict], drops: list[dict], order_number: int, dt_lkp: dict) -> list[
        JobStep]:
        dt_this = dt_lkp[order_number]
        inc = 1
        steps = []
        for load in loads:
            steps.append(
                PlatformScienceOrderSyncStep._convert_load_or_drop_to_job_step(load, inc, order_number, "Load",
                                                                               dt_this))
            inc += 1
        for drop in drops:
            steps.append(
                PlatformScienceOrderSyncStep._convert_load_or_drop_to_job_step(drop, inc, order_number, "Drop",
                                                                               dt_this))
            inc += 1
        return steps

    @staticmethod
    def _convert_task(eid: str, product_name: str, task_order: int, load_or_drop: Literal["Load", "Drop"],
                      completed: bool, completed_at: datetime) -> JobTask:
        return JobTask(
            remarks=[],
            fields={},
            id=str(task_order),
            name=f"{load_or_drop} Product ({product_name})",
            # What types are available and how do we use them? "arrived" is the only one I have verified so far
            type="arrived",
            completed=completed,
            completed_at=completed_at,
            external_id=eid,
            order=task_order,
            status="New",
        )

    @staticmethod
    def _convert_load_or_drop_to_job_step(data: dict, step_order: int, order_number: int,
                                          load_or_drop: Literal["Load", "Drop"], dt_this: dict) -> JobStep:
        product_names = {x["product_name"] for x in data["details"]}
        # Every detail that we care about can be represented by the first item in its group
        detail_len = len(product_names)
        detail_slug = f"{detail_len} product" + ("s" if detail_len > 1 else "")

        completed = data["status"] == "complete"
        completed_at = None
        # Locate the update in the DT data, and filter to route events for this order specifically
        for route in filter(lambda x: x["order_number"] == order_number, dt_this["route"]):
            action_map = {
                "Load": "loading",
                "Drop": "dropping"
            }
            action_matches = action_map[load_or_drop] == route["action"]
            destination = route["destination_id"]
            if action_matches and destination == data["location_id"]:
                completed = route["complete"]
                if completed:
                    completed_at = route["end_time"]
                else:
                    completed_at = None
                break

        return JobStep(
            tasks=[
                PlatformScienceOrderSyncStep._convert_task(f"{order_number}-{step_order}-{task_order}", d, task_order,
                                                           load_or_drop, completed, completed_at)
                for task_order, d in enumerate(product_names)
            ],
            order=step_order,
            completed=completed,
            completed_at=completed_at,
            type="New",
            name=f"{load_or_drop} {detail_slug}",
            external_id=f"{order_number}-{step_order}",
            location_external_id=data["location_id"]
        )


if __name__ == "__main__":
    async def main():
        step = PlatformScienceOrderSyncStep(
            sd_client=GravitateSDAPI(
                base_url="",
                client_id="",
                client_secret=""
            ),
            psc=PlatformScienceClient(
                base_url="",
                client_id="",
                client_secret=""
            ),
            order_nums=[],
            mongo_database=MongoClient("mongodb conn str")["db_name"]
        )
        res = await step.execute(None)


    asyncio.run(main())



import asyncio

from loguru import logger

from bb_integrations_lib.protocols.pipelines import Step


class SleepStep(Step):
    def __init__(self, duration: float, *args, **kwargs):
        """
        Equivalent to ``asyncio.sleep(duration)``.

        :param duration: The number of seconds to sleep.
        """
        super().__init__(*args, **kwargs)

        self.duration = duration

    def describe(self) -> str:
        return f"Wait a specified number of seconds"

    async def execute(self, data: None) -> None:
        logger.debug(f"Sleeping for {self.duration} seconds...")
        await asyncio.sleep(self.duration)



import itertools
from datetime import datetime
from operator import attrgetter
from typing import Callable

from bb_integrations_lib.models.dtn_supplier_invoice import DTNSupplierInvoice, Item, ItemTax, SummaryTax, \
    DeferredTaxItem
from bb_integrations_lib.protocols.pipelines import Step
from bb_integrations_lib.shared.model import SDSupplierInvoiceCreateRequest, SDSupplierInvoiceDetail, \
    SDSupplierInvoiceDetailType


class ConvertDTNInvoiceToSDModel(Step):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def describe(self) -> str:
        return "Convert DTN invoice data model to SD supplier invoice data model"

    async def execute(self, i: DTNSupplierInvoice) -> SDSupplierInvoiceCreateRequest:
        return self.convert_dtn_invoice(i)

    def convert_dtn_invoice_deferred_tax_item(self, deferred_tax: DeferredTaxItem) -> SDSupplierInvoiceDetail:
        return SDSupplierInvoiceDetail(
            bol_number=None, # Not present for summary taxes, since they're invoice level
            bol_date_local=None,
            type=SDSupplierInvoiceDetailType.tax,
            product=None,
            tax_type=None,
            tax_description=deferred_tax.description,
            tax_authority=None,
            tax_non_deferred=False,
            rate=deferred_tax.amount,
            amount=1,
            total=deferred_tax.amount,
            uom=None,
            gross_volume=None,
            net_volume=None,
        )

    def convert_dtn_invoice_summary_tax(self, summary_tax: SummaryTax) -> SDSupplierInvoiceDetail:
        return SDSupplierInvoiceDetail(
            bol_number=None, # Not present for summary taxes, since they're invoice level
            bol_date_local=None,
            type=SDSupplierInvoiceDetailType.tax,
            product=None,
            tax_type=summary_tax.tax_code,
            tax_description=summary_tax.description,
            tax_authority=None,
            tax_non_deferred=not summary_tax.deferred,
            rate=summary_tax.rate,
            amount=summary_tax.quantity_billed,
            total=summary_tax.line_total,
            uom=summary_tax.unit_of_measure,
            gross_volume=None,
            net_volume=None,
        )

    def convert_dtn_invoice_item_tax(
            self, item_tax: ItemTax, item_product: str, bol_date: datetime | None
    ) -> SDSupplierInvoiceDetail:
        return SDSupplierInvoiceDetail(
            bol_number=item_tax.bol_number,
            bol_date_local=bol_date,
            type=SDSupplierInvoiceDetailType.tax,
            product={"source_name": item_product},
            tax_type=item_tax.tax_code,
            tax_description=item_tax.description,
            tax_authority=None,
            tax_non_deferred=not item_tax.deferred,
            rate=item_tax.rate,
            amount=item_tax.quantity_billed,
            total=item_tax.line_total,
            uom=item_tax.unit_of_measure,
            gross_volume=item_tax.quantity_billed,
            net_volume=item_tax.quantity_billed,
        )

    def convert_dtn_invoice_item(self, invoice_detail: Item) -> list[SDSupplierInvoiceDetail]:
        product = invoice_detail.description
        bol_date = invoice_detail.ship_datetime
        item_detail = SDSupplierInvoiceDetail(
            bol_number=invoice_detail.bol_number,
            bol_date_local=bol_date,
            type=SDSupplierInvoiceDetailType.supply,
            product={"source_name": product},
            rate=invoice_detail.rate,
            amount=invoice_detail.quantity_billed,
            total=invoice_detail.line_total,
            uom=invoice_detail.unit_of_measure,
            gross_volume=invoice_detail.gross_quantity or invoice_detail.quantity_billed,
            net_volume=invoice_detail.net_quantity or invoice_detail.quantity_billed
        )
        tax_details = [
            self.convert_dtn_invoice_item_tax(tax, product, bol_date) for tax in invoice_detail.tax_records or []
        ]
        # Should we support RINS numbers here? How?
        return [item_detail] + tax_details

    @staticmethod
    def resolve_field(invoice: DTNSupplierInvoice, attr_getter: Callable[[Item], str]) -> str | None:
        """
        Resolve a specific field from the invoice line items, if they are all the same.
         :return: The return value retrieved by calling attr_getter on each invoice item, if all are the same (or if
           only one item is present), or None if no items are present or if any items disagree with each other.
        """
        # Get a list of terminals from the invoice line items, where the terminal is not blank
        item_fields = list(map(attr_getter, invoice.items))
        if len(set(item_fields)) == 1:
            return item_fields[0]
        return None

    def convert_dtn_invoice(self, invoice: DTNSupplierInvoice) -> SDSupplierInvoiceCreateRequest:
        deferred_tax_details = []
        summary_tax_details = []
        if invoice.deferred_taxes is not None:
            deferred_tax_details.extend([
                self.convert_dtn_invoice_deferred_tax_item(detail_item)
                for detail_item in invoice.deferred_taxes.items
            ])
        for summary_tax in invoice.summary_taxes:
            summary_tax_details.append(
                self.convert_dtn_invoice_summary_tax(summary_tax)
            )
        return SDSupplierInvoiceCreateRequest(
            invoice_number=invoice.header.invoice_number,
            source_name=invoice.header.sold_to_name,
            supplier=invoice.header.seller_name,
            terminal=self.resolve_field(invoice, attrgetter("ship_from_name")) or None,
            due_date_local=invoice.header.invoice_due_date,
            # Convert date to datetime at midnight
            invoice_date_local=datetime.combine(invoice.header.invoice_date, datetime.min.time()),
            # Get a flattened list, since each call to convert_dtn_invoice_item yields a list of 1 or more detail items
            # Then tack on summary and deferred tax details, when applicable.
            details=list(itertools.chain.from_iterable(
                [self.convert_dtn_invoice_item(item) for item in invoice.items]
            )) + summary_tax_details + deferred_tax_details,
            extra_data={
                "invoice_filename": invoice.original_filename
            } if invoice.original_filename else {},
            ship_to_city=self.resolve_field(invoice, attrgetter("ship_to_city")) or None,
            ship_to_state=self.resolve_field(invoice, attrgetter("ship_to_state")) or None,
            ship_from_city=self.resolve_field(invoice, attrgetter("ship_from_city")) or None,
            ship_from_state=self.resolve_field(invoice, attrgetter("ship_from_state")) or None,
        )



import json
from datetime import datetime, UTC, timedelta
from typing import Dict, Any, Tuple, Iterable

import polars as pl
import pytz
from loguru import logger

from bb_integrations_lib.models.pipeline_structs import StopPipeline
from dateutil.parser import parse

from bb_integrations_lib.protocols.pipelines import Step
from bb_integrations_lib.shared.exceptions import StepInitializationError
from bb_integrations_lib.shared.model import RawData, FileFormat


class ReadingParser:
    """
    Parser for tank reading data that supports multiple output formats.

    This class provides parsing capabilities for tank reading data into different
    client-specific formats, including standard PDI-compatible output and various
    Circle K formats for different integration systems.

    The parser takes preprocessed DataFrame input (with standardized column names)
    and transforms it according to the specified format requirements.

    Supported formats:
        - standard: PDI-compatible format with configurable disconnection tracking
        - circlek: Circle K format with TelaPoint integration fields
        - circlek2: Simplified Circle K format for Gravitate system integration
    """

    def __init__(self, file_format: FileFormat):
        """
        Initialize the parser with a specific output format.

        Args:
            file_format (FileFormat): The desired output format for parsed data
        """
        self.format = file_format

    def parse(self, df: pl.DataFrame, disconnected_column: bool = False,
              disconnected_only: bool = False, water_level_column: bool = False) -> pl.DataFrame:
        """
        Parse tank reading data according to the configured format.

        Args:
            df (DataFrame): Input DataFrame with standardized column names:
                - Store Number: Store identifier
                - Name: Store name
                - Tank Id: Tank identifier
                - Tank Product: Product type in tank
                - Carrier: Carrier information
                - Volume: Current volume measurement
                - Ullage: Unfilled space in tank
                - Read Time: Timestamp of reading
                - Store Source Number: Store number assigned by client  (extra_data.site_source_number)
                - Disconnected: Boolean disconnection status (optional)
            disconnected_column (bool): Whether to include Disconnected column
                in output (standard format only)
            disconnected_only (bool): Whether to filter to only disconnected
                tanks (standard format only)
            water_level_column (bool): Whether to include Water Level column

        Returns:
            DataFrame: Parsed data in the specified format

        Raises:
            ValueError: If the configured format is not supported
        """
        if self.format == FileFormat.standard:
            return self._parse_standard(df, disconnected_column, disconnected_only, water_level_column)
        elif self.format == FileFormat.circlek:
            return self._parse_circlek(df)
        elif self.format == FileFormat.circlek2:
            return self._parse_circlek2(df)
        elif self.format == FileFormat.reduced:
            return self._parse_reduced(df)
        else:
            raise ValueError(f"Unsupported format: {self.format}")

    def _parse_standard(self, df: pl.DataFrame, disconnected_column: bool,
                        disconnected_only: bool, water_level_column: bool) -> pl.DataFrame:
        """
        Parse data into standard PDI-compatible format.

        Produces columns: Store Number, Name, Tank Id, Tank Product, Carrier,
        Volume, Ullage, Read Time, and optionally Disconnected.
        """
        column_order = [
            'Store Number', 'Name', 'Tank Id', 'Tank Product',
            'Carrier', 'Volume', 'Ullage', 'Read Time'
        ]

        if water_level_column:
            column_order.append('Water')

        if disconnected_column:
            column_order.append('Disconnected')

        if disconnected_only:
            df = df.filter(pl.col("Disconnected") == True)

        return df.select(column_order)

    def _parse_circlek(self, df: pl.DataFrame) -> pl.DataFrame:
        """
        Parse data into Circle K format with TelaPoint integration fields.

        Transforms input data into Circle K's expected structure with TelaPoint
        account/site numbers and formatted timestamps.
        """

        def parse_date(dt: str) -> str:
            dt = parse(dt)
            return dt.strftime("%m/%d/%Y %H:%M")

        rows = []
        records = df.to_dicts()

        for record in records:
            rows.append({
                "ClientName": None,
                "FacilityName": None,
                "FacilityInternalID": None,
                "FacilityState": None,
                "VolumePercentage": None,
                "TankStatus": None,
                "TankNbr": None,
                "TankInternalID": None,
                "AtgTankNumber": record['Tank Id'],
                "ATGTankLabel": None,
                "Product": None,
                "TankCapacity": None,
                "Ullage": None,
                "SafeUllage": None,
                "Volume": record['Volume'],
                "Height": None,
                "Water": None,
                "Temperature": None,
                "InventoryDate": parse_date(record['Read Time']),
                "SystemUnits": None,
                "CollectionDateTimeUtc": None,
                "TelaPointAccountNumber": 100814,
                "TelaPointSiteNumber": record['Store Number'],
            })

        return pl.DataFrame(rows)

    def _parse_circlek2(self, df: pl.DataFrame) -> pl.DataFrame:
        """
        Parse data into simplified Circle K format.
        """

        def parse_date(dt: str) -> str:
            dt = parse(dt)
            return dt.strftime("%m/%d/%Y %H:%M")

        rows = []
        records = df.to_dicts()

        for record in records:
            rows.append({
                "storeNumber": record.get('Store Source Number'),
                "timestamp": parse_date(record['Read Time']),
                "tankLabel": record.get('Tank Product'),  # Product name assigned to tank
                "volume": record['Volume'],
                "tankNumber": record['Tank Id'],
                "ullage": record.get('Ullage', 0),
                "productLevel": 0,  # Can be set to 0 as specified
                "waterLevel": 0,  # Can be set to 0 as specified
                "temperature": 0  # Can be set to 0 as specified
            })

    def _parse_reduced(self, df: pl.DataFrame) -> pl.DataFrame:
        """
        Parse data into reduced format with minimal columns.
        """
        df = df.with_columns([
            (pl.col('Read Time')
            .str.replace(" UTC", "")
            .str.to_datetime(format="%Y-%m-%d %H:%M:%S %z")
            .dt.convert_time_zone("UTC")
            .dt.strftime("%Y-%m-%dT%H:%M:%S.%3f")
            .str.replace(r"\.(\d{3})\d*$", ".$1")
            + pl.lit("Z"))
            .alias('read_time')
        ])

        return df.select([
            pl.col('Store Number').alias('store_number'),
            pl.col('Tank Id').alias('tank_id'),
            pl.col('read_time'),
            pl.col('Volume').alias('volume')
        ])

    @classmethod
    def create_parser(cls, file_format: FileFormat) -> 'ReadingParser':
        """
        Factory method to create a parser for the specified format.

        Args:
            file_format (FileFormat): The desired output format

        Returns:
            ReadingParser: Configured parser instance
        """
        return cls(file_format)


class ParseTankReadingsStep(Step):
    def __init__(self, file_format: FileFormat, timezone: str, include_water_level: bool = False,
                 disconnected_column: bool = False, disconnected_only: bool = False,
                 disconnected_hours_threshold: float | None = None,
                 *args, **kwargs):
        """
        Parse tank readings from BBDExportReadingsStep and create a PDI-compatible file, either in the standard output
        format (with some configuration options) or in a client-specific format.

        See the ``FileFormat`` enum for currently supported formats.

        :param file_format: The ``FileFormat`` to use for the output file. "standard" provides a PDI compatible file,
          but additional formats may be implemented at client request.
        :param timezone: Timezone to localize read times to. Must be a Pytz-known timezone name.
        :param include_water_level: Whether to include water level in the output file. Defaults to False.
        :param disconnected_column: Whether to include a "Disconnected" column in the output file. Independent of
          disconnected_only. Requires ``disconnected_hours_threshold`` to be set.
        :param disconnected_only: Whether to post-filter result rows to only disconnected site/tanks. Independent of
          disconnected_column. Requires ``disconnected_hours_threshold`` to be set.
        :param disconnected_hours_threshold: How long it may be since the last reading before a tank is considered
          disconnected. Setting this value without ``disconnected_column`` or ``disconnected_only`` will have no effect.
        """
        super().__init__(*args, **kwargs)
        self.file_format: FileFormat = file_format
        self.timezone = timezone
        self.step_created_time = datetime.now(UTC)
        self.include_water_level = include_water_level
        self.disconnected_column = disconnected_column
        self.disconnected_only = disconnected_only
        self.disconnected_hours_threshold = disconnected_hours_threshold
        self.disconnected_threshold = timedelta(
            hours=self.disconnected_hours_threshold) if self.disconnected_hours_threshold else None

        # Initialize the reading parser for the configured format
        self.reading_parser = ReadingParser(self.file_format)

        if (self.disconnected_column or self.disconnected_only) and not self.disconnected_hours_threshold:
            raise StepInitializationError(
                "If disconnected_column or disconnected_only is True, disconnected_hours_threshold must be set")

    def describe(self) -> str:
        return f"Format tank readings step"

    async def execute(self, data: Tuple[Dict, Dict, Iterable]) -> pl.DataFrame:
        store_lkp, tank_lkp, readings = data
        df = pl.LazyFrame(readings, schema={
            "tank_agent_name": str,
            "store_number": str,
            "run_time": datetime,
            "tank_id": str,
            "read_time": datetime,
            "product": str,
            "monitor_type": str,
            "volume": float,
        })
        return (await self.parse_data(data=df, tank_lkp=tank_lkp, store_lkp=store_lkp)).collect()

    @staticmethod
    def safe_expand_extra_data(df: pl.DataFrame, extra_data_col='extra_data') -> pl.DataFrame:
        if extra_data_col not in df.columns:
            logger.warning(f"Warning: {extra_data_col} column not found")
            return df
        if df[extra_data_col].is_null().all():
            logger.warning(f"Warning: All {extra_data_col} values are null")
            return df.drop(extra_data_col)
        try:
            col_dtype = df[extra_data_col].dtype
            if str(col_dtype).startswith("Struct"):
                df = df.unnest(extra_data_col)
            else:
                df = df.with_columns([
                    pl.col(extra_data_col).map_elements(
                        lambda x: x if isinstance(x, dict) else (json.loads(x) if isinstance(x, str) else {}),
                        return_dtype=pl.Struct
                    ).alias(f"{extra_data_col}_parsed")
                ])
                df = df.unnest(f"{extra_data_col}_parsed")
                df = df.drop(extra_data_col)
            return df
        except Exception as e:
            logger.error(f"Error expanding {extra_data_col}: {e}")
            return df

    def maybe_add_water_level(self, df: pl.DataFrame, columns_to_keep: list) -> pl.DataFrame:
        if self.include_water_level:
            columns_to_keep.append('water')
            df = df.with_columns([
                pl.col('water').fill_nan(None).alias('water')
            ])
        return df

    def map_stores_to_gravitate_name(self, df: pl.DataFrame, store_lkp: Dict) -> pl.DataFrame:
        store_name_map = {k: v.get('name') if v else None for k, v in store_lkp.items()}
        store_source_number_map = {k: v.get('extra_data', {}).get('site_source_number') if v else None
                                   for k, v in store_lkp.items()}

        df = df.with_columns([
            pl.col('store_number').replace(store_name_map, default=None).alias('name'),
            pl.col('store_number').replace(store_source_number_map, default=None).alias('store_source_number')
        ])
        return df

    def select_keep_columns(self, df: pl.DataFrame, columns_to_keep: list) -> pl.DataFrame:
        return df.select(columns_to_keep)

    def localize_timestamps(self, df: pl.DataFrame) -> pl.DataFrame:
        df = df.with_columns([
            pl.col('read_time')
            .dt.replace_time_zone("UTC")
            .dt.convert_time_zone(self.timezone)
            .alias('read_time')
        ])
        return df

    def maybe_calculate_disconnected_tanks(self, df: pl.DataFrame) -> pl.DataFrame:
        """
        Optionally calculate disconnected status for tanks.

        Groups by (store_number, tank_id) and determines if tank is disconnected
        based on reading timestamps and configured threshold.

        Args:
            df: Input DataFrame with read_time column

        Returns:
            DataFrame with optional 'disconnected' column added
        """
        if self.disconnected_column or self.disconnected_only:
            disconnections = df.group_by(['store_number', 'tank_id']).agg([
                pl.col('read_time').map_elements(
                    lambda times: self.is_disconnected(times, self.disconnected_threshold),
                    return_dtype=pl.Boolean
                ).first().alias('disconnected')
            ])
            df = df.join(disconnections, on=['store_number', 'tank_id'], how='left')
        return df

    def format_timestamps(self, df: pl.DataFrame) -> pl.DataFrame:
        """
        Format read_time datetime to string using vectorized strftime.

        Converts to "YYYY-MM-DD HH:MM:SS TZ±HHMM" format.

        Args:
            df: Input DataFrame with datetime read_time column

        Returns:
            DataFrame with string-formatted read_time
        """
        df = df.with_columns([
            pl.col('read_time').dt.strftime("%Y-%m-%d %H:%M:%S %Z%z").alias('read_time')
        ])
        return df

    def create_tank_lookup_key(self, df: pl.DataFrame) -> pl.DataFrame:
        df = df.with_columns([
            (pl.col('store_number').cast(pl.Utf8) + ':' + pl.col('tank_id').cast(pl.Utf8)).alias('key')
        ])
        return df

    def map_tank_metadata(self, df: pl.DataFrame, tank_lkp: Dict) -> pl.DataFrame:
        """
        Map tank metadata (product, carrier, storage_max) using composite key.
        """
        tank_product_map = {k: v.get('product') if v else None for k, v in tank_lkp.items()}
        tank_carrier_map = {k: v.get('carrier') if v else None for k, v in tank_lkp.items()}
        tank_storage_max_map = {k: v.get('storage_max') if v else None for k, v in tank_lkp.items()}
        df = df.with_columns([
            pl.col('key').replace(tank_product_map, default=None).alias('tank_product'),
            pl.col('key').replace(tank_carrier_map, default=None).alias('carrier'),
            pl.col('key').replace(tank_storage_max_map, default=None).alias('storage_max')
        ])
        return df

    def calculate_ullage(self, df: pl.DataFrame) -> pl.DataFrame:
        """
        Formula: ullage = storage_max - volume
        """
        df = df.with_columns([
            (pl.col('storage_max').fill_null(0) - pl.col('volume').fill_null(0)).alias('ullage')
        ])
        return df

    def map_tanks_to_gravitate_id(self, df: pl.DataFrame, tank_lkp: Dict) -> pl.DataFrame:
        df = self.create_tank_lookup_key(df)
        df = self.map_tank_metadata(df, tank_lkp)
        df = self.calculate_ullage(df)
        return df

    def format_column_names(self, df: pl.DataFrame) -> pl.DataFrame:
        column_mapping = {col: ParseTankReadingsStep.format_column_name(col) for col in df.columns}
        return df.rename(column_mapping)

    def add_reference_data(self, df: pl.DataFrame, tank_lkp: Dict, store_lkp: Dict) -> pl.DataFrame:
        columns_to_keep = ['store_number', 'tank_id', 'read_time', 'volume', 'name', 'store_source_number']
        df = self.maybe_add_water_level(df, columns_to_keep)
        df = self.map_stores_to_gravitate_name(df, store_lkp)
        df = self.select_keep_columns(df, columns_to_keep)
        df = self.localize_timestamps(df)
        df = self.maybe_calculate_disconnected_tanks(df)
        df = self.format_timestamps(df)
        df = self.map_tanks_to_gravitate_id(df, tank_lkp)
        df = self.format_column_names(df)

        return df

    def file_parser(self, df: pl.DataFrame, tank_lkp: Dict, store_lkp: Dict) -> pl.DataFrame:
        """
        Parse tank reading data into the configured output format.

        This method first enriches the data with reference information, then uses
        the ReadingParser to transform it into the appropriate format.

        Args:
            df (DataFrame): Raw tank readings DataFrame
            tank_lkp (Dict): Tank lookup dictionary
            store_lkp (Dict): Store lookup dictionary

        Returns:
            DataFrame: Parsed DataFrame in the configured output format
        """
        df = self.add_reference_data(df, tank_lkp, store_lkp)
        return self.reading_parser.parse(
            df,
            disconnected_column=self.disconnected_column,
            disconnected_only=self.disconnected_only,
            water_level_column=self.include_water_level
        )

    def localize(self, dt: datetime, timezone: str = None) -> datetime:

        if timezone is None:
            timezone = self.timezone
        utc = pytz.timezone('UTC')
        dt = utc.localize(dt)
        dt = dt.astimezone(pytz.timezone(timezone))
        return dt

    def format_dt_col(self, dt: datetime) -> str:
        return dt.strftime("%Y-%m-%d %H:%M:%S %Z%z")

    def is_disconnected(self, reading_times: Iterable[datetime], threshold: timedelta) -> bool:
        """
        Determine if a tank is disconnected based on reading timestamps.

        A tank is considered disconnected if it has no readings within the threshold
        period from the current time, or if it has no readings at all.

        Args:
            reading_times (Iterable[datetime]): Collection of reading timestamps
            threshold (timedelta): Time threshold for disconnection detection

        Returns:
            bool: True if tank is disconnected, False otherwise
        """
        # Skip future times, with a 15 minute grace period (maybe the clocks are just slightly desynced)
        filtered = [t for t in reading_times if t <= self.step_created_time + timedelta(minutes=15)]
        # No readings, or a reading older than now - threshold? Disconnected
        return len(filtered) == 0 or max(filtered) < datetime.now(UTC) - threshold

    @staticmethod
    def format_column_name(col_name: str) -> str:
        return ' '.join(word.capitalize() for word in col_name.split('_'))

    async def parse_data(self, data: pl.DataFrame, store_lkp: Dict, tank_lkp: Dict) -> pl.DataFrame:
        """
        Parse tank reading data and return as RawData for pipeline output.

        Args:
            data (DataFrame): Raw tank readings DataFrame
            store_lkp (Dict): Store lookup dictionary
            tank_lkp (Dict): Tank lookup dictionary

        Returns:
            DataFrame: Parsed data output in the requested format. An additional step is required to export this data
            to a RawData object.
        """
        df = self.file_parser(df=data, tank_lkp=tank_lkp, store_lkp=store_lkp)
        # self.pipeline_context.included_files["parse pricing engine prices to PDI file step"] = json.dumps(
        #     df.to_dicts())
        return df



from typing import Any

from bb_integrations_lib.protocols.pipelines import Step, Input, StepConfig
from loguru import logger

from bb_integrations_lib.provider.api.cargas.client import CargasClient
from bb_integrations_lib.provider.api.cargas.model import CreateWholesaleTicketRequest, \
    CreateWholesaleTicketRequestBundle


class CargasWholesaleBundleUploadStep(Step):
    def __init__(self, cargas_client: CargasClient, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.cargas_client = cargas_client

    def describe(self):
        return "Upload a Cargas Wholesale ticket request bundle"

    async def execute(self, i: CreateWholesaleTicketRequestBundle) -> None:
        cwt_resp = await self.cargas_client.create_wholesale_ticket(i.ticket_request)
        doc_id = cwt_resp["ResponseValues"]["DocumentID"]

        # Now try to upload each line
        for index, line in enumerate(i.line_requests):
            try:
                line.DocumentID = int(doc_id)
                await self.cargas_client.create_wholesale_line(line)
            except Exception as e:
                logger.error(f"Failed to upload line {index} to Cargas: {e}")
        print(cwt_resp)

        # TODO: Upload fee line items



import warnings

from bb_integrations_lib.provider.gcp.cloud_secrets.client import CloudSecretsClient

class GCSMClient(CloudSecretsClient):
    def __init__(self):
        warnings.warn("Use provider.gcp.cloud_secrets.client.CloudSecretsClient", DeprecationWarning, stacklevel=2)
        super().__init__()



from typing import Any, Dict, Optional, Self
from enum import Enum
from pydantic import BaseModel
from datetime import datetime, UTC

class ReferenceDataType(str, Enum):
    terminal = "terminal"
    store = "store"
    tank = "tank"
    location = "location"
    product = "product"
    counterparty = "counterparty"


class CoreMasterReferenceData(BaseModel):
    """
    A single 'item' in the industry. This could point to a store, a product, a company, or anything else.
    The data here lives in the RITA master tenant only, and works in conjunction with the data stored in RITA client
    tenant's tables to enable conversion from one client's representation to another.
    """
    data_type: ReferenceDataType
    """The data type of this item. Used for lookups."""

    mrid: str
    """The unique ID of this item. MasterReferenceLink objects will link on this field."""

    name: str
    """Display name of this item."""

    source_tenant: str | None = None
    """The original tenant of this item, if it has one."""

    matching_info: Dict[str, Any] = {}
    """A dictionary describing the info that makes this object unique. E.g. for terminals this could be the federal TCN."""
    # We plan to engineer some way to throw a client's data at an LLM for matching. Matching_info would be the place to store
    # data that's similar in structure.

    mrd_extra_data: Dict[str, Any] = {}
    """Any additional data about the object that should be tied to it during the conversion process."""

    updated_by: Optional[str] = None
    updated_on: Optional[datetime] = datetime.now(UTC)


class MasterReferenceData(CoreMasterReferenceData):
    children: Dict[str, CoreMasterReferenceData] = {}
    """The children of this master reference data"""


class CoreMasterReferenceLink(BaseModel):
    """
    Links 'items' in this tenant's gravitate instance to the MasterReferenceData. The data here lives only in RITA client
    tenants, not the RITA master tenant, and matches some record in the MasterReferenceData. It is possible for one
    MasterReferenceLink to match one MasterReferenceData.
    """
    data_type: ReferenceDataType
    """The data type of this item. Used for lookups."""

    mrid: Optional[str] = None
    """Linked MasterReferenceData item. This may be empty and indicates that there is no associated item."""

    display_name: str
    """
    Display name of this item, in the context of the tenant's gravitate. This is not used for code lookups and can be
    set to anything. It may be set automatically by the sync module, if the source_system is "Gravitate"
    """

    source_id: str
    """
    Either (but not both): 
      - Mongodb ID of the item in the tenant's gravitate (if source_system is "Gravitate").
      - ID of the item in the specified source system.
    """

    matching_info: Dict[str, Any] = {}
    """A dictionary describing the info that makes this object unique. E.g. for terminals this could be the federal TCN, or an address."""

    mrl_extra_data: Dict[str, Any] = {}
    """Any additional data about the object that should be tied to it during the conversion process."""


    updated_by: Optional[str] = None
    updated_on: Optional[datetime] = datetime.now(UTC)


class MasterReferenceLink(CoreMasterReferenceLink):
    source_system: str = "Gravitate"
    """The source system of this item. Defaults to 'Gravitate'. Used in mapping requests to add detail in mappings."""

    children: Dict[str, CoreMasterReferenceLink] = {}
    """Child references keyed by their Source ID. These can only match child references of the parent's matched master
    reference data."""


class ReferenceDataMappingExtraData(BaseModel):
    origin_mrl_extra_data: dict[str, Any] = {}
    mrd_extra_data: dict[str, Any] = {}


class ReferenceDataMapping(BaseModel):
    origin_source_id: str = ""
    origin_source_system: str = ""
    origin_mrl_id: str | None = None
    origin_tenant: str = ""
    target_tenant: str | None = None
    target_mrls: list[MasterReferenceLink] = []
    matched_mrid: str | None = None
    matched_child_mrid: str | None = None
    extra_data: ReferenceDataMappingExtraData = ReferenceDataMappingExtraData()
    milliseconds_taken: int | None = None
